{==============================================================================|
| Project : Notes/Delphi class library                           | 3.9.1       |
|==============================================================================|
| Content:                                                                     |
|==============================================================================|
| The contents of this file are subject to the Mozilla Public License Ver. 1.0 |
| (the "License"); you may not use this file except in compliance with the     |
| License. You may obtain a copy of the License at http://www.mozilla.org/MPL/ |
|                                                                              |
| Software distributed under the License is distributed on an "AS IS" basis,   |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for |
| the specific language governing rights and limitations under the License.    |
|==============================================================================|
| Initial Developers of the Original Code are:                                 |
|   Sergey Kolchin (Russia) skolchin@yahoo.com                                 |
|   Sergey Kucherov (Russia)                                                   |
|   Sergey Okorochkov (Russia)                                                 |
| All Rights Reserved.                                                         |
|   Last Modified:                                                             |
|     25.08.01                                                                 |
|==============================================================================|
| Contributors and Bug Corrections:                                            |
|   Fujio Kurose                                                               |
|   Noah Silva                                                                 |
|   Tibor Egressi                                                              |
|   Andreas Pape                                                               |
|   Anatoly Ivkov                                                              |
|   Winalot                                                                    |
|   Dmitry Mokrushin                                                           |
|   Olaf Hahnl                                                                 |
|     and others...                                                            |
|==============================================================================|
| History: see README.TXT                                                      |
|==============================================================================|
| Main library unit                                                            |
|                                                                              |
| Use NO_NATIONAL_TEXT define to enable/disable national text processing       |
| Use NO_CTRL_BREAK to disable Control-Break handling                          |
|                                                                              |
| Latest versions are available at                                             |
|   http://www.geocities.com/skolchin/                                         |
|==============================================================================|}
unit Class_LotusNotes;


{.$DEFINE NO_NATIONAL_TEXT}
{.$DEFINE NO_CTRL_BREAK}
{.$DEFINE NO_INIT_SECTION}

{$IFNDEF WIN32}
-- This unit is for Windows 32 environment
{$ENDIF}
{$RANGECHECKS OFF}
{$ALIGN OFF}

{$IFDEF NO_INIT_SECTION}
{.$WEAKPACKAGEUNIT ON}
{$ENDIF}


interface

uses
  Messages, SysUtils, Classes, Windows, System.Variants, Util_LNApi,
  System.Generics.Collections, System.Types, System.AnsiStrings;

const
  strLength = 2048;
  ITEM_VALUE_SEPARATOR = ';';
  LOOKUP_VALUE_END = #9'==='#9;

type
  TNotesItem = class;
  TNotesDocumentCollection = class;
  TNotesDocument = class;
  TNotesDatabase = class;
  TNotesACL = class;
  TNotesACLEntry = class;
  PTDateTime = ^TDateTime;
  TNotesItemClass = class of TNotesItem;

  ELotusNotes = class(Exception)
    // All LN exceptions
    ErrorCode: DWORDapi;
    constructor CreateErr (aCode: DWORDapi; aMsg: string);
  end;

  ELnFormulaCompile = class(ELotusNotes)
    // Formula compilation
    CompileErrCode: DWORDapi;
    CompileErrMsg: string;
    CompileErrOffset, CompileErrLength: word;
    constructor CreateErr (aCode: DWORDapi; aMsg: string; aErrCode: DWORDapi; aErrOff,aErrLen: word);
  end;

  ELnInvalidFolder = class(ELotusNotes)
    // Attempt to open a view as a folder
  end;

  // Document collection
  TNotesDocumentCollection = class(TList)
  private
    Db: TNotesDatabase;
    fUnreadDocs: boolean;

    function  GetDocumentId(Idx: integer): longint;
    function  GetDocument(Idx: integer): TNotesDocument;
    procedure GetCollectionByHandle(hBuffer: LHANDLE; NotesFound: word; fSummary: boolean);
    function GetSummaryValues (Idx: integer): string;
  public
    constructor Create(notesDatabase: TNotesDatabase);
    destructor Destroy; override;

    procedure Clear; override;

    // Use with caution!
    function AddDocumentId (ID: longint; Summary: string): integer;
    procedure DeleteDocumentID (Index: integer);

    property Database: TNotesDatabase read Db;
    property Document[Ind: integer]: TNotesDocument read GetDocument;
      // Returns NEW document object based on stored ID.
      // Don't forget to free it!
    property DocumentId[Ind: integer]: longint read GetDocumentId; default;

    procedure MarkAllRead (fRead: boolean);
      // Sets all documents read/unread
    property SummaryValues[Idx: integer]: string read GetSummaryValues;
  end;

  // Notes View (basic class)
  TNotesView = class(TNotesDocumentCollection)
  private
    FNoteID: NOTEID;
    Flags: string;
    FKey: string;
    FMaxDocs: integer;
    function GetIsFolder: boolean;
    function GetIsShared: boolean;
    function GetUniversalID: UNID;
    function GetOriginatorID: OID;
  protected
    FName: string;
    constructor CreateEmpty (notesDatabase: TNotesDatabase; aName: string);
    class function LoadFlags (Database: TNotesDatabase; anID: NOTEID): string;
    procedure Load(aKey: string; nMaxDocs: integer);
    procedure LoadByID(anID: NOTEID; aKey: string; nMaxDocs: integer);
  public
    constructor Create(notesDatabase: TNotesDatabase; aName: string);
    constructor CreateExt(notesDatabase: TNotesDatabase; aName: string; aKey: string; nMaxDocs: integer);

    class function OpenView (notesDatabase: TNotesDatabase; aName: string): TNotesView;
      // use this function to create object of valid view/folder class

    class function OpenViewExt(notesDatabase: TNotesDatabase; aName: string; aKey: string; nMaxDocs: integer): TNotesView;
      // the same as OpenView, but allows to limit view selection by the key or documents count

    procedure Update;

    class function GetViewID (notesDatabase: TNotesDatabase; aName: string): NOTEID;

    property Name:     string read FName;
    property ID: NOTEID read FNoteID;
    property UniversalID: UNID read GetUniversalID;
    property OriginatorID: OID read GetOriginatorID;
    property IsFolder: boolean read GetIsFolder;
    property IsShared: boolean read GetIsShared;
    property Key:      string  read FKey;
    property MaxDocs:  Integer read FMaxDocs;
  end;

  // Notes folder
  TNotesFolder = class(TNotesView)
  private
    procedure SetName(Value: string);
  public
    constructor CreateNew (notesDatabase: TNotesDatabase; aName: string;
      fShared: boolean; FormatFolder: TNotesFolder);

    procedure AddDocument (Doc: TNotesDocument);
    procedure AddDocuments (DocList: TList);

    // replaces old Clear method
    procedure RemoveDocuments;

    function  Copy (NewName: string): TNotesFolder;
    procedure Delete;
    procedure DeleteDocument (Doc: TNotesDocument);
    procedure DeleteDocuments (DocList: TList);
      //WARNING! Free the folder object right after calling this function
    procedure Move (ParentFolder: TNotesFolder);

    property Name: string read FName write SetName;
  end;

  // Document item
  TNotesItem = class(TPersistent)
  private
    fDoc: TNotesDocument;
    fName: string;
    fItemBid: BlockId;
    fValueBid: BlockId;
    fItemflags: word;
    fDataType: word;
    fSaveToDisk: boolean;
    fValueLength: DWORDapi;
    fIsNew: boolean;
    fCreated: boolean;
    FStringsValue: TStrings;
    FSeqNo: integer;
    function GetAsFloat : Extended;
    function GetAsList  : Variant;
    function GetAsNumbers : Variant;
    function GetAsString : String;
    procedure InitItemInfo(PrevItem: BLOCKID);
    procedure SetItemFlags (anItemFlag: integer; anValue: boolean);
    function  GetItemFlags (anItemFlag: integer): boolean;
    function  GetLastModifed: TDateTime;
    function  GetAsDateTime: TDateTime;
    function GetAsReference: UNID;
    function GetAsStrings : TStrings;
    function GetAsTimes : variant;
    procedure SetAsDateTime(Value: TDateTime);
    function  GetItemValue: variant;
    procedure SetAsFloat (Value: extended);
    procedure SetAsList (Value: variant);
    procedure SetAsNumbers (Value: variant);
    procedure SetAsReference(Value: UNID);
    procedure SetAsString (Value: string);
    procedure SetAsStrings (Value: TStrings);
    procedure SetAsTimes (Value: variant);
    procedure SetItemValue (aValue: variant);
    procedure SetCreated;
  protected
    function GetRichText: TStrings; virtual;
    procedure SetRichText (Value: TStrings); virtual;
  public
    constructor Create(notesDocument: TNotesDocument; aName: string); virtual;
    constructor CreateNew (notesDocument: TNotesDocument; aName: string); virtual;
    constructor CreateNext (notesItem: TNotesItem);
    destructor Destroy; override;

    // Item flags
    property IsAuthors    : boolean index ITEM_READWRITERS  read GetItemFlags write SetItemFlags;
    property IsEncrypted  : boolean index ITEM_SEAL         read GetItemFlags write SetItemFlags;
    property IsNames      : boolean index ITEM_NAMES        read GetItemFlags write SetItemFlags;
    property IsProtected  : boolean index ITEM_PROTECTED    read GetItemFlags write SetItemFlags;
    property IsReaders    : boolean index ITEM_READERS      read GetItemFlags write SetItemFlags;
    property IsSigned     : boolean index ITEM_SIGN         read GetItemFlags write SetItemFlags;
    property IsSummary    : boolean index ITEM_SUMMARY      read GetItemFlags write SetItemFlags;
    property IsNewItem    : boolean read fCreated;
    property ItemFlags    : word read fItemflags write fItemFlags;

    property LastModified : TDateTime read GetLastModifed;
    property Name: string read fName;
    property SaveToDisk: boolean read fSaveToDisk write fSaveToDisk;
    property Document: TNotesDocument read fDoc;

    property ItemType: word read fDataType write fDataType; //see TYPE_... in Util_NotesAPI
    property ValueLength: DWORDapi read fValueLength;
    property Value: Variant read GetItemValue write SetItemValue;

    property AsString: string read GetAsString write SetAsString;
    property AsDateTime: TDateTime read GetAsDateTime write SetAsDateTime;
    property AsNumber: extended read GetAsFloat write SetAsFloat;
    property AsList: variant read GetAsList write SetAsList;
    property AsNumbers: variant read GetAsNumbers write SetAsNumbers;
    property AsStrings: TStrings read GetAsStrings write SetAsStrings;
    property AsTimes: variant read GetAsTimes write SetAsTimes;
    property AsRichText: TStrings read GetRichText write SetRichText;
    property AsReference: UNID read GetAsReference write SetAsReference;

    // Low-level value access- don't use
    property ItemBid: BLOCKID read fItemBid;
    property ValueBid: BLOCKID read fValueBid;
    procedure GetValueBuffer(var Buffer: pointer; var BufSize: DWORDapi);
    procedure SetValueBuffer(iType,wFlags: word;  Buffer: pointer;  BufSize: DWORDapi);

    // Multiple items support
    function CreateNextItem: TNotesItem;
    function NextItemExists: boolean;
    function LoadNextItem: boolean;
    property SeqNo: integer read FSeqNo;
  end;

  // Notes document
  TNotesDocument = class(TPersistent)
  private
    fHandle: LHandle;
    fId: NOTEID;
    FFields: TStringList;
    FDatabase: TNotesDatabase;
    FAttach: TStrings;
    FFontTable: TStrings;
    FIsRead: boolean;
    FProfileName: string;
    FSummary: TStrings;
    function GetAttachment(Index: integer): string;
    function GetAttachmentCount: integer;
    function GetRecipients: string;
    procedure SetRecipients(Value: string);
    function GetItemByNum (ItemNum: integer): string;
    procedure SetItemByNum (ItemNum: integer; Value: string);
    function GetBodyAsString: string;
    procedure SetBodyAsString(Value: string);
    function GetBodyAsMemo: TStrings;
    function GetFieldCount : longint;
    function GetFieldName (Index: integer): string;
    procedure SetBodyAsMemo(Value: TStrings);
    function GetItemByName (ItemName: string): TNotesItem;
    procedure SetIsRead (Value: boolean);
    function GetLastModified: TDateTime;
    function GetLastAccessed: TDateTime;
    function GetUniversalID: UNID;
    function GetOriginatorID: OID;
    procedure InitDocument(notesDatabase: TNotesDatabase;  anID: DWORDapi);
  public
    FMaxAttachment: word; // for RTI, don't use
    FMaxFontID: word;

    constructor CreateNew(notesDatabase: TNotesDatabase);
      // create new DOCUMENT in the database (see also TNotesDatabase.CreateDocument)
    constructor Create (notesDatabase: TNotesDatabase; anId: longint);
      // open a document by its NoteID
    constructor CreateByUNID(notesDatabase: TNotesDatabase;  anUNID: UNID);
      // open a document by its Universal ID
    constructor CreateProfile (notesDatabase: TNotesDatabase; aProfileName: string);
      // opens a profile document
    constructor CreateResponse(notesDatabase: TNotesDatabase;  MainDoc: TNotesDocument);
      // creates a new response document
    constructor CreateResponseByUNID(notesDatabase: TNotesDatabase; const anUNID: UNID);
      // creates a new response document with UNID only - Subject is not set!
    destructor Destroy; override;

    // These are Document attachment handling
    property  AttachmentCount: integer read GetAttachmentCount;  //number of file attachments
    property  Attachment[Index: integer]: string read GetAttachment;
    function  Attach (AName: string): integer; //attach a file by its name
    procedure Detach (Index: integer; FileName: string);
      //detach a file attached to a form with a given name
    function  FindAttachment(aName: string): integer;  //return index in Attachment or -1
    procedure DeleteAttachment(Index: integer);   //delete an attachment
    procedure SaveAttachment(Index: integer; FilePath: String);

    procedure DeleteItem (ItemName: string);  //deletes an item
    function IsItemExists(ItemName: string): boolean;
    function ReplaceItemValue(itemName: string; value: Variant): TNotesItem; // sets field value
    procedure Save(force: boolean = False; createResponse: boolean = False; markRead: boolean = False); // saves a document
    procedure Send(fAttachForm: boolean = False; ARecipients: string = '');
      // sends a document. if ARecipients <> '', overrides previously assigned addressees
      // if Database.SaveMail = True, also saves a document
    procedure CopyItem (Source: TNotesDocument; itemName: string);
    procedure AttachForm (aForm: string);
    function ComputeWithForm (doDataTypes, raiseError : boolean) : boolean;

    function GetSignature (var SignedBy: string; var CertifiedBy: string; pTime: PTDateTime): boolean;
      // Get signature information. SignedBy is user name, CertifiedBy - domain name. pTime can be nil
      // Returns True if document is signed and signature is valid
    procedure Sign; //signs a document
    procedure CheckAddress;
      // Checks names assigned to SendTo and Recipients fields agains server address book

    function CopyToDatabase(DestDB:TNotesDatabase):TNotesDocument;
      //Added by Sergey Okorochkov

    function Evaluate(aFormula: string): variant;
      // Evaluates a formula on the document

    // Note properties
    property DocID: NOTEID read fID;
    property Handle: LHandle read fHandle;
    property Database: TNotesDatabase read fDatabase;
    property LastModified: TDateTime read GetLastModified;
    property LastAccessed: TDateTime read GetLastAccessed;
    property UniversalID: UNID read GetUniversalID;
    property OriginatorID: OID read GetOriginatorID;
    property ProfileName: string read FProfileName;
      // blank for non-profile documents

    // Special fields for mail documents
    // Note that using this fields do not change doc items obtained from Items property
    property Form: string index MAIL_FORM_ITEM_NUM read GetItemByNum write SetItemByNum; //form name
    property Subject: string index MAIL_SUBJECT_ITEM_NUM read GetItemByNum write SetItemByNum; //subject field
    property SendTo: string index MAIL_SENDTO_ITEM_NUM read GetItemByNum write SetItemByNum;
    property CopyTo: string index MAIL_COPYTO_ITEM_NUM read GetItemByNum write SetItemByNum;
    property BlindCopyTo: string index MAIL_BLINDCOPYTO_ITEM_NUM read GetItemByNum write SetItemByNum;
      // Addresses, multiple ones are separated by ','.
    property MailFrom: string index MAIL_FROM_ITEM_NUM read GetItemByNum write SetItemByNum; // sender
    property Recipients: string read GetRecipients write SetRecipients;
      // List of recipients. Don't change manually, it's generated from SendTo + CopyTo + BlindCopyTo
      // If set, these properties are ignored
    property Body: TStrings read GetBodyAsMemo write SetBodyAsMemo;   // body field
    property BodyAsString: string read GetBodyAsString write SetBodyAsString;

    // Items
    // Use Items[Name].As... to read/write document properties
    property FieldCount: longint read GetFieldCount;
    property FieldName[Index: integer]: string read GetFieldName;
    property Items[ItemName: string]: TNotesItem read GetItemByName; default;

    // Multiple items support
    function CountMultipleItems(aName: string): integer;
        //returns a number of items with given name
    function LoadMultipleItems(aName: string): TList;
        //loads all items with given name and returns a list of them
        //the resulting list contains item objects, which must be free'd
        //manually, along with the list itself

    // Summary values are available only when a document was read from view (TNotesView class)
    // This class contains text representations of all values placed in summary buffer upon save
    // It's not updated during document saving
    property SummaryValues: TStrings read FSummary;

    // Clears fields list and reloads it from the document
    // Call Save before to commit changes
    procedure ReloadFields;
    procedure UpdateUnread;
    property IsRead: boolean read FIsRead write SetIsRead;

    // Font table access, internal use
    property FontTable: TStrings read FFontTable;
    procedure LoadFontTable;
    procedure SaveFontTable;
  end;

  // Notes database
  TNotesDatabase = class
  private
    FACL: TNotesACL;
    FHandle: integer;
    FFileName: string;
    FServerName: string;
    FSaveMail: boolean;
    FViews: TStrings;
    function GetACL: TNotesACL;
    function GetFullName: string;
    function GetActive: boolean;
    function GetDatabaseID: DBID;
    function GetReplicaInfo: DBReplicaInfo;
    procedure SetActive(Value: boolean);
    function GetTitle: string;
    function GetViewByIndex (Index: integer): TNotesView;
    function GetViews (ViewName: string): TNotesView;
    function GetViewCount: integer;
    procedure SetTitle(Value: string);
    procedure SetFileName(Value: string);
    procedure SetReplicaInfo(Value: DBReplicaInfo);
    procedure SetServerName(Value: string);
    function GetQuotaInfo: DBQUOTAINFO;
  protected
    procedure CheckViews;
    procedure UpdateViews;
  public
    constructor Create; // this doesn't open or create a database!
    destructor Destroy; override;

    function CreateDocument: TNotesDocument;
      // creates new empty document in the database
      // use Form property to assign the form
    function CreateResponseDocument (const aResUNID: UNID; aSubject: string): TNotesDocument;
      // creates a new response for given document. See also CreateResponse constructors

    //Moved by Sergey Okorochkov
    //Changed by KOL - 09.03.99
    function FindNotes (formula: string; notesDateTime: TDateTime; noteClass: word; fSummary: boolean): TNotesDocumentCollection;

    function Open (aServer, dbFile: string): boolean;
      // open specified database. Use blank server name to open from disk
    procedure OpenMail;
      // Open mailbox. Equal to Open (MailServer, MailFile);
    procedure Close;
      // close the database
    procedure CloseSession;
      // close the database and break a connection
    procedure CopyRecords(SourceDB: TNotesDatabase);
      // copies all records from SourceDB
    procedure CreateNew(aServer, dbFile: string; TemplateDB: string);
      // creates a new database with given name. TemplateDB is full name of template or ''
    procedure SendMail (Address: string; Subject: string; Body: string);
      // Simple mail sending
    procedure ReplyMail(Mail: TNotesDocument;  Body: string);
      // Simple reply
    procedure Delete(DocID: integer); //delete a document from database
    procedure DeleteDocument(Doc: TNotesDocument);
      // Delete an open document. Use this function if a document is akredy opened
      // WARNING! After deleting, you cannot access document properties!

    procedure ListAddressBooks (aServer: string; List: TStrings; fGetTitles: boolean);
      // list AB for specific server or for local system if aServer=""
      // if fGetTitles=False, List will contain only paths to address books
      // otherwise, it will be <path>=<title>
    procedure OpenPrivateAddressBook;
      // open local address book

    function FTSearch(query: string; maxDocs, sortOptions, otherOptions: integer): TNotesDocumentCollection;
      // Search by Full-text index
      // maxDocs limits returned docs number (now unused)
      // for sortOptions and otherOptions look in Notes documentation
    function FindDocument(Formula: string): TNotesDocument;
      // finds one document

    function Search(formula: string; notesDateTime: TDateTime = 0; maxDocs: integer = 0): TNotesDocumentCollection;
      // searches for documents by given formula.
      // notesDateTime limits documents by creation date (since ...)
      // maxDocs limits returned docs number (now unused)
    function UnreadDocuments : TNotesDocumentCollection;
      // Returns collection of all unread documents in a database
    procedure MarkRead (NoteID: DWORDapi; fRead: boolean);
      // Changes given note status. Use MarkAllRead to process many docs
    procedure MarkAllRead (Docs: TNotesDocumentCollection; fRead: boolean);
      // Marks all docs read/unread

    // Views
    property ViewCount: integer read GetViewCount;
    property ViewByIndex[Index: integer]: TNotesView read GetViewByIndex;
    property Views[ViewName: string]: TNotesView read GetViews;
      // Don't free objects!
    property ViewNames: TStrings read FViews;

    function OpenView(AName: string; notesDateTime: TDateTime; maxDocs: integer): TNotesView;
      // obsolete - use Views property instead
      // must free returned object
      // notesDateTime and maxDocs are unused

  	// ACL
    property ACL: TNotesACL read GetACL;
      // Returns ACL for this database
      // If ACL doesn't exist it will be created
    procedure CopyACL(SourceDb: TNotesDatabase);
      //copies an ACL of a database to this one

    property Handle: integer read fHandle; //handle for direct access
    property SaveMail: boolean read FSaveMail write FSaveMail; //if true, mailed docs are also saved in the database
    property FileName: string read FFileName write SetFileName; //db file name
    property Server: string read FServerName write SetServerName; //db server name
    property Active: boolean read GetActive write SetActive;
    property FullName: string read GetFullName;
    property Title: string read GetTitle write SetTitle;  //database title

    class function UserName: string;        //name of logged user
    class function LmbcsUserName: string;   //name of logged user as LMBSC string
    class function MailFileName: string;    //name of mailbox database
    class function MailServer: string;      //mail server

    function NotesVersion: word;
    property DatabaseID: DBID read GetDatabaseID;
    property ReplicaInfo: DBReplicaInfo read GetReplicaInfo write SetReplicaInfo;

    // Quotas
    property QuotaInfo: DBQuotaInfo read GetQuotaInfo;
  end;

  // One directory entry
  TNotesDirEntry = record
    FileName: array [0..MAXPATH-1] of AnsiChar;
    EntryType: boolean;                              //False for file, True for directory
    FileInfo: array [0..MAXPATH-1] of AnsiChar;      //description of the database
  end;
  pNotesDirEntry = ^TNotesDirEntry;

  // Search options
  // nfoFiles and nfoTemplates are mutually exclusive. If both are set, nfoFiles is used
  TNotesFindOption = (nfoFiles, nfoTemplates, nfoSubDirs);
  TNotesFindOptions = set of TNotesFindOption;

  //This class allows to list servers/directories/databases}
  TNotesDirectory = class
  private
    FPorts: TStrings;
    hDirectory: THandle;
    SrcServer: string;
    SrcPath: string;
    SrcOptions: TNotesFindOptions;
    SrcTable: TList;
    SrcIndex: integer;

    function GetPorts: TStrings;
  public
    constructor Create;
    destructor Destroy; override;

    procedure FindClose;
    // closes search sequence

    function FindFirst (const Server, Path: string; Options: TNotesFindOptions; var Entry: TNotesDirEntry): boolean;
    function FindNext (var Entry: TNotesDirEntry): boolean;
      // use these functions to find files in the directory
      // they returns False if no more files can be found

    procedure ListServers (Port: string; List: TStrings);
      // lists all servers for specific port (if empty string - for all ports)
      // at least one entry always exists: empty string for local computer

    property Ports: TStrings read GetPorts;
      // available ports
  end;

  // Options for LookupName funcion
  TNotesLookupOption = (
    nloAll,        //returns all entries of given category
    nloNoSearch,   //look only in the first AB
    nloExhaustive  //search in all AB on the server
  );
  TNotesLookupOptions = set of TNotesLookupOption;

  {Lotus Notes name parsing}
  TNotesName = class
  private
    FName: string;  //canonical
    FComponents: DN_COMPONENTS;
    fParsed: boolean;
    FTemplateName: string;

    function GetAbbreviatedName: string;
    function GetKeyword : string;
    procedure SetAbbreviatedName (Value: string);
    procedure SetName (Value: string);
  public
    constructor Create (aName: string);
      // The name must be in either canonical or abbreviated format
      // You can use function IsCanonical to determine name type

    class function IsCanonical (aName: string): boolean;
    function IsHirerarchical : boolean;
      // Returns True if the name is in canonical distinguished format

    class function TranslateName (aName: string; fToCanonical: boolean; aTemplate: string): string;
      // Translation function. If flag=True, tries to canonilize, otherwise - abbreviate

    function GetNameComponent (Index: integer): string;
    procedure SetNameComponent (Index: integer; Value: string);
      // Common name parts access functions

    property Common:     string index 0 read GetNameComponent write SetNameComponent;
    property Given:      string index 1 read GetNameComponent write SetNameComponent;
    property Surname:    string index 2 read GetNameComponent write SetNameComponent;
    property Initials:   string index 3 read GetNameComponent write SetNameComponent;
    property Generation: string index 4 read GetNameComponent write SetNameComponent;
    property Country:    string index 5 read GetNameComponent write SetNameComponent;
    property OrgUnit1:   string index 6 read GetNameComponent write SetNameComponent;
    property OrgUnit2:   string index 7 read GetNameComponent write SetNameComponent;
    property OrgUnit3:   string index 8 read GetNameComponent write SetNameComponent;
    property OrgUnit4:   string index 9 read GetNameComponent write SetNameComponent;
    property ADMD:       string index 10 read GetNameComponent write SetNameComponent;
    property PRMD:       string index 11 read GetNameComponent write SetNameComponent;
      // these two are mutually exclusive
    property Organization: string index 12 read GetNameComponent write SetNameComponent;

    property Abbreviated: string read GetAbbreviatedName write SetAbbreviatedName;
    property Canonical: string read FName write SetName;
    property Keyword: string read GetKeyword;
    property TemplateName: string read FTemplateName write FTemplateName;
      // template name used if a name contains no parts except user name (like 'Sergey Kolchin/')

    // ****************
    // Lookup functions
    class function LookupName (ServerName: string; NameSpaces: array of string;
      Names: array of string; Items: array of string; Flags: TNotesLookupOptions;
      Values: TStrings): boolean;
      // Generic function for name lookup in the address book (AB)
      // ServerName defines location of AB ('' for local)
      // NameSpaces define view in AB. Use USER_NAMESPACE or '' for default
      // Names is an array of names to look up. Names may = [''] if Flags specifies NAME_LOOKUP_ALL
      // Items is an array of item names to return with each match (cannot be [''])
      // Flags defines options for lookup
      // The function returns True on success and fills Values with all matched values found
      // Note that the function can handle only text items

    class function LookupNameList (ServerName: string; NameSpaces: TStrings;
      Names: TStrings; Items: TStrings; Flags: TNotesLookupOptions;
      Values: TStrings): boolean;
      // The same as above, but takes lists as parameters

    class function CheckAddress (aServer, aName: string): boolean;
      // Simple way to check mail address
      // Checks for existence of the name in the AB on the server
  end;

  // ACL
  TNotesACL = class
  private
    FDatabase: TNotesDatabase;
    FEntries: TList;
    FHandle: LHandle;
    FRoles: TStrings;
    function GetEntriesCount: integer;
    function GetEntry (aName: string): TNotesACLEntry;
    function GetEntryByIndex (Index: integer): TNotesACLEntry;
    function GetUniformAccess: boolean;
    procedure ReadEntries;
    procedure SetUniformAccess (Value: boolean);
  public
    constructor Create (aDatabase: TNotesDatabase);
    destructor Destroy; override;
    procedure Save;

    function CreateACLEntry (aName: string): TNotesACLEntry;
    procedure DeleteACLEntry (aName: string);
    property EntriesCount: integer read GetEntriesCount;
    property Entry[aName: string]: TNotesACLEntry read GetEntry; default;
    property EntryByIndex[Index: integer]: TNotesACLEntry read GetEntryByIndex;

    property Database: TNotesDatabase read FDatabase;
    property Handle: LHandle read FHandle;
    property Roles: TStrings read FRoles;
    property UniformAccess: boolean read GetUniformAccess write SetUniformAccess;
  end;

  // ACL access levels and flags
  TNotesAclAccessLevel = (
    aclNoAccess,
    aclDepositor,
    aclReader,
    aclAuthor,
    aclEditor,
    aclDesigner,
    aclManager
  );

  TNotesAclFlag = (
    acfAuthorNoCreate,        //Authors can't create new notes (only edit existing ones) }
    acfServer,                //Entry represents a Server (V4) }
    acfNoDelete,              //User cannot delete notes }
    acfCreatePersonalAgents,  //User can create personal agents (V4) }
    acfCreatePersonalFolders, //User can create personal folders (V4) }
    acfPerson,                //Entry represents a Person (V4) }
    acfGroup,                 //Entry represents a group (V4) }
    acfCreateFolders,         //User can create and update shared views & folders (V4)
    acfCreateLotusScript,     //User can create LotusScript }
    acfPublicReader,          //User can read public notes }
    acfPublicWriter,          //User can write public notes }
    acfAdminReaderAuthor,     //Admin server can modify reader and author fields in db }
    acfAdminServer            //Entry is administration server (V4) }
  );
  TNotesAclFlags = set of TNotesAclFlag;

  // ACL entry
  TNotesACLEntry = class
  private
    FAccessLevel: TNotesAclAccessLevel;
    FACL: TNotesACL;
    FFlags: TNotesAclFlags;
    FName, FOldName: string;
    FPrivileges: ACL_PRIVILEGES;
    FNew: boolean;
    FUpdateFlags: word;

    constructor Create(anACL: TNotesACL; aName: string; AccLevel: TNotesAclAccessLevel;
      const AclPrivs: ACL_PRIVILEGES; AccFlags: TNotesAclFlags);
    function GetRoles: string;
    procedure SetAccessLevel (Value: TNotesAclAccessLevel);
    procedure SetFlags (Value: TNotesAclFlags);
    procedure SetName (Value: string);
    procedure SetPrivileges (Value: ACL_PRIVILEGES);
    procedure SetRoles (Value: string);
  public
    procedure AddRole (aRole: string);
    constructor CreateNew(anACL: TNotesACL; aName: string;
                          AccLevel: TNotesAclAccessLevel; aRoles: string);
    procedure DeleteRole (aRole: string);
    destructor Destroy; override;

    procedure Update;

    property AccessLevel: TNotesAclAccessLevel read FAccessLevel write SetAccessLevel;
    property ACL: TNotesACL read FACL;
    property Flags: TNotesAclFlags read FFlags write SetFlags;
    property Name: string read FName write SetName;
    property Privileges: ACL_PRIVILEGES read FPrivileges write SetPrivileges;
    property Roles: string read GetRoles write SetRoles;
  end;

// Thread initialization
// InitNotesThread must be called by the thread before using of any Notes functions
// CloseNotesThread must be called by the thread before thread terminates
procedure InitNotesThread;
procedure CloseNotesThread;

// Date time conversions
function DateTimeToNotes (DelphiTime: TDateTime): TIMEDATE;
function NotesToDateTime (NotesTime: TIMEDATE): TDateTime;
function NotesToDateTimeEx (NotesTime: TimeStruct): TDateTime;

// UNID conversion
{$IFDEF D3}
function UNIDtoStr (const anUNID: UNID; fDelimiters: boolean): string;
{$ELSE}
function UNIDtoStr (const anUNID: UNID; fDelimiters: boolean = True): string;
{$ENDIF}

// Text conversion
// Work only if LN_NATIONAL_TEXT is defined
// Written by Fujio Kurose(fujio.kurose@nifty.ne.jp)
function Lmbcs2Native (aString: string): string;    overload;
function Lmbcs2Native (aString: AnsiString): AnsiString; overload;
function Native2Lmbcs (aString: AnsiString): AnsiString;  //-- overload;


// String conversions
//   Converts Notes-style string (where #0 marks line-feeds) to Delphi
function NotesToString(NotesStr: PAnsiChar; StrSize: integer): string;
//   Opposite conversion. NotesStr is allocated internally and MUST BE FREED by FreeMem()
//   Returns True if NotesStr has embedded zeroes
function StringToNotes(DelphiStr: string; var NotesStr: PAnsiChar; var StrSize: integer): boolean;

// Color mapping
// TColor is not used because it's declared in Graphics unit
function NotesToColor(NotesColor: integer): integer;
function ColorToNotes(DelphiColor: integer): integer;

// File paths...
function ConstructPath (const aServer, aPath: string): string;
procedure ParsePath (const aPath: string; var aServer, aFile: string);
function GetNotesExeDir: string;
function GetNotesDataDir: string;
function GetNotesIniFile: string;

// Adds a large item to note by separating it to multiple items with the same name
procedure HugeNSFItemAppend(hNote: NOTEHANDLE;
                          ItemFlags: Word;
                          Name: PAnsiChar;
                          NameLength: Word;
                          DataType: Word;
                          Value: Pointer;
                          ValueLength: LongInt);

// Control-break signal handler function (internal use)
function NDBreakProc: STATUS; far; stdcall;


implementation
uses Util_LNApiErr, Class_NotesRTF;

(******************************************************************************)
// Exceptions
(******************************************************************************)
constructor ELotusNotes.CreateErr (aCode: DWORDapi; aMsg: string);
begin
  inherited Create(aMsg);
  ErrorCode := aCode;
end;

(******************************************************************************)
constructor ELnFormulaCompile.CreateErr(aCode: DWORDapi; aMsg: string; aErrCode: DWORDapi; aErrOff,aErrLen: word);
var
 buf: AnsiString;
begin
 setLength(buf, 255);
 OSLoadString(0, aErrCode, PAnsiChar(buf), 254);
 buf := Lmbcs2Native(System.AnsiStrings.StrPas(PAnsiChar(buf)));
 inherited CreateErr (aCode, format ('%s: %s (at %d:%d)',[aMsg, buf, aErrOff,aErrLen]));
 CompileErrCode := aErrCode;
 CompileErrMsg := String(buf);
 CompileErrOffset := aErrOff;
 CompileErrLength := aErrLen;
end;

(******************************************************************************)
// Init and callbacks
(******************************************************************************)
procedure InitNotesThread;
begin
  CheckError (NotesInitThread);
end;

(******************************************************************************)
procedure CloseNotesThread;
begin
  NotesTermThread;
end;

(******************************************************************************)
function  NSFItemScanProc(Spare, ItemFlags: word;
                          Name: PAnsiChar;
                          NameLength: word;
                          Value: pointer;
                          ValueLength: DWORDapi;
                          RoutineParameter: pointer): STATUS;  stdcall;
begin
 TStringList(RoutineParameter).Add(
                       String(Lmbcs2Native(Copy(System.AnsiStrings.strPas(Name), 1, NameLength))));
 //TStringList (RoutineParameter).Add(Copy (System.AnsiStrings.strPas(name), 1, NameLength));
 Result := NOERROR;
end;

(******************************************************************************)
// Conversion procedures
(******************************************************************************)
function DateTimeToNotes (DelphiTime: TDateTime): TIMEDATE;
var
  T: TimeStruct;
  x,y,z,hs: word;
begin
  //OSCurrentTimeDate (@T.GM);
  //TimeGMToLocal (T);
  FillChar (T, sizeOf(T), #0);
  DecodeDate (DelphiTime,x,y,z);
  T.Year := x;
  T.Month := y;
  T.Day := z;
  DecodeTime(DelphiTime,x,y,z,hs);
  T.hour := x;
  T.minute := y;
  T.second := z;
  T.hundredth := hs div 10;
  //T.weekday := DayOfWeek (DelphiTime);
  OSCurrentTimeZone (@T.zone, @T.dst);
  TimeLocalToGM(T);
  Result := T.GM;
end;

(******************************************************************************)
function NotesToDateTime (NotesTime: TIMEDATE): TDateTime;
var
  T: TimeStruct;
begin
  T.GM := NotesTime;
  OSCurrentTimeZone (@T.zone, @T.dst);
  Result := NotesToDateTimeEx (T);
end;

(******************************************************************************)
{function NotesToDateTimeEx (NotesTime: TimeStruct): TDateTime;
begin
  TimeGMToLocal(NotesTime);
  Result := 0;
  if NotesTime.year>0 then Result := EncodeDate(NotesTime.year,NotesTime.month,NotesTime.day);
  if NotesTime.hour>0 then Result := Result + EncodeTime(NotesTime.hour,NotesTime.minute,
    NotesTime.second,NotesTime.hundredth);
end;}

(******************************************************************************)
function NotesToDateTimeEx (NotesTime: TimeStruct): TDateTime;
var
   failed: boolean;
begin
   Result := 0;   //default

   //Try this first, but note that dates without a time are undefined
   //(from API reference)
   failed := TimeGMToLocalZone (NotesTime);

   //If TimeGMToLocalZone failed, use TimeGMToLocal instead
   if failed then failed := TimeGMToLocal (NotesTime);

   //One of the above was successful, so convert to TDateTime
   if not failed then try
      Result := EncodeDate (NotesTime.year, NotesTime.month, NotesTime.day);
      if NotesTime.hour = -1 then NotesTime.hour := 0;
      if NotesTime.minute = -1 then NotesTime.minute := 0;
      if NotesTime.second = -1 then NotesTime.second := 0;
      if NotesTime.hundredth = -1 then NotesTime.hundredth := 0;
      Result := Result + EncodeTime (NotesTime.hour, NotesTime.minute,
                                     NotesTime.second, NotesTime.hundredth);
   except
   end;
end;

var
  OldBreakProc: OSSIGBREAKPROC;

(******************************************************************************)
function NDBreakProc: STATUS; far; stdcall;
begin
  if (GetAsyncKeyState(VK_CANCEL) and 1) = 0 
    then Result := NOERROR
    else Result := ERR_CANCEL;
end;

(******************************************************************************)
// Translation
(******************************************************************************)
function Lmbcs2Native (aString: string): string;
var
  n: DWORDapi;
  ansiStr, ansiResult: AnsiString;
begin
{$IFNDEF NO_NATIONAL_TEXT}
  ansiStr := AnsiString(aString);
  ansiResult := '';
  n := length(ansiStr);
  SetLength(ansiResult, n + 2);
  n := OSTranslate(
    OS_TRANSLATE_LMBCS_TO_NATIVE,
    PAnsiChar(ansiStr),
    n,
    PAnsiChar(ansiResult),
    n);
  ansiResult[n+1] := #0;
  Result := String(ansiResult);
{$ELSE}
  Result := aString;
{$ENDIF}
end;

function Lmbcs2Native (aString: AnsiString): AnsiString;
var
  n: DWORDapi;
begin
{$IFNDEF NO_NATIONAL_TEXT}
  n := length(aString);
  Result := '';
  SetLength(Result, n + 2);
  n := OSTranslate(
    OS_TRANSLATE_LMBCS_TO_NATIVE,
    PAnsiChar(AString),
    n,
    PAnsiChar(Result),
    n);
  Result[n+1] := #0;
  Result := System.AnsiStrings.StrPas(PAnsiChar(Result));
{$ELSE}
  Result := aString;
{$ENDIF}
end;

function Native2Lmbcs (aString: AnsiString): AnsiString;
var
  n: DWORDapi;
begin
{$IFNDEF NO_NATIONAL_TEXT}
  n := length(aString) * 3; //LMBCS is ~3 times bigger - Fujio Kurose
  Result := '';
  SetLength(Result, n + 2);
  n := OSTranslate(
    OS_TRANSLATE_NATIVE_TO_LMBCS,
    PAnsiChar(AString),
    length(aString),
    PAnsiChar(Result),
    n);
  Result[n+1] := #0;
  Result := System.AnsiStrings.StrPas(PAnsiChar(Result));
{$ELSE}
  Result := aString;
{$ENDIF}
end;

(******************************************************************************)
// Misc
(******************************************************************************)
function ConstructPath (const aServer, aPath: string): string;
var
  ansiServer, ansiPath, ansiResult: AnsiString;
begin
  ansiServer := AnsiString(aServer);
  ansiPath   := Native2Lmbcs(AnsiString(aPath));
  setLength (ansiResult, 256);
  CheckError (OSPathNetConstruct('', PAnsiChar(ansiServer), PAnsiChar(ansiPath), PAnsiChar(ansiResult)));
  Result := String(System.AnsiStrings.StrPas(PAnsiChar(Lmbcs2Native(ansiResult))));
end;

(******************************************************************************)
procedure ParsePath (const aPath: string; var aServer, aFile: string);
var
  ansiPort: AnsiString;
  ansiServer, ansiPath, ansiFile: AnsiString;
begin
  ansiServer := AnsiString(aServer);
  ansiPath   := Native2Lmbcs(AnsiString(aPath));
  ansiFile   := AnsiString(aFile);
  setLength (ansiServer, 256);
  setLength (ansiFile, 256);
  setLength (ansiPort, 256);
  CheckError (OSPathNetParse(PAnsiChar(ansiPath),PAnsiChar(ansiPort),PAnsiChar(ansiServer),PAnsiChar(ansiFile)));
  aServer := String(System.AnsiStrings.StrPas(PAnsiChar(Lmbcs2Native(ansiServer))));
  aFile   := String(System.AnsiStrings.StrPas(PAnsiChar(Lmbcs2Native(ansiFile))));
end;

(******************************************************************************)
function GetNotesExeDir: string;
var
 ansiResult: AnsiString;
begin
  setlength(ansiResult,MAXPATH+1);
  OSGetExecutableDirectory (PAnsiChar(ansiResult));
  Result := String(System.AnsiStrings.StrPas(PAnsiChar(ansiResult)));
  if (Result <> '') and (Result[length(Result)] <> '\') then
    Result := Result + '\';
end;

(******************************************************************************)
function GetNotesIniFile: string;
var
 ansiResult: AnsiString;
begin
  setlength(ansiResult,MAXPATH+1);
  OSGetIniFileName (PAnsiChar(ansiResult));
  Result := String(System.AnsiStrings.StrPas(PAnsiChar(ansiResult)));
end;

(******************************************************************************)
// By Andy
function GetNotesDataDir: string;
var
 ansiResult: AnsiString;
begin
  setLength(ansiResult, MAXPATH+1);
  if OSGetEnvironmentString( 'Directory', PAnsiChar(ansiResult), MAXPATH) then
    Result := String(System.AnsiStrings.StrPas(PAnsiChar(ansiResult)))
  else
    Result := '';
  if (Result <> '') and (Result[length(Result)] <> '\') then
    Result := Result + '\';
end;


(******************************************************************************)
// Buffer-to-value conversion
(******************************************************************************)
function BufferToValue(Buffer: pointer; BufLen: DWORDapi): Variant;
var
 ItemType, nEntry, i, wLen: word;
 str: string;
 num: NUMBER;
 tm:  TIMEDATE;
 pc:  PAnsiChar;
 PRnge: PRANGE;
 PNumValue: PNUMBER;
 PDtValue: PTIMEDATE;
 buf: PAnsiChar;
begin
 Result := NULL;
 ItemType := pword(Buffer)^;
 Buffer := pointer (DWORDapi(Buffer) + sizeOf(word));
 dec (BufLen, sizeOf(word));
 case ItemType of
   TYPE_TEXT: begin
     {str := '';
      setLength (str, BufLen+1);
      //Move (PAnsiChar(Buffer)[0], PAnsiChar(str)[0], BufLen);
      System.AnsiStrings.strLCopy (PAnsiChar(str), PAnsiChar(Buffer), BufLen);
      str[BufLen+1] := #0;
      Result := System.AnsiStrings.(PAnsiChar(str));}
     GetMem(buf, BufLen+1);
     try
       Move(Buffer^, buf^, BufLen);
       buf[BufLen] := #0;
       Result := Lmbcs2Native(NotesToString(buf, BufLen));
     finally
       FreeMem(buf);
     end;
   end;
   TYPE_TEXT_LIST: begin
     nEntry := ListGetNumEntries(Buffer, False);
     if nEntry > 1 then Result := VarArrayCreate([0, nEntry-1], varOleStr);
     for i := 0 to nEntry-1 do begin
       ListGetText (Buffer, False, i, @pc, @wLen);
        {setLength (str, wLen + 2);
        System.AnsiStrings.strLCopy (PAnsiChar(str), pc, wLen);
        str[wLen+1] := #0;
        if nEntry > 1
          then Result[i] := Lmbcs2Native(System.AnsiStrings.(PAnsiChar(str)))
          else Result := Lmbcs2Native(System.AnsiStrings.(PAnsiChar(str)));}
       GetMem(buf, wLen+1);
       try
         Move(pc^, buf^, wLen);
         buf[wLen] := #0;
         str := Lmbcs2Native(NotesToString(buf, wLen));
         if nEntry > 1
           then Result[i] := Lmbcs2Native(str)
           else Result := Lmbcs2Native(str);
       finally
         FreeMem(buf);
       end;
      end;
    end;
    TYPE_NUMBER: begin
      num := PNUMBER(Buffer)^;
      Result := num;
    end;
    TYPE_NUMBER_RANGE: begin
      PRnge := PRANGE (Buffer);
      PNumValue := PNUMBER (DWORDapi(Buffer) + sizeOf(USHORT)*2);
      if PRnge^.ListEntries = 0 then Result := Null
      else if PRnge^.ListEntries = 1 then Result := PNumValue^
      else begin
        Result := VarArrayCreate ([0, PRnge^.ListEntries-1], varDouble);
        for i := 0 to PRnge^.ListEntries-1 do begin
          Result[i] := PNumValue^;
          PNumValue := PNUMBER (DWORDapi(PNumValue) + sizeof(NUMBER));
        end;
      end;
    end;

    TYPE_TIME: begin
      tm := PTIMEDATE(Buffer)^;
      Result := NotesToDateTime (tm);
    end;

    TYPE_TIME_RANGE: begin
      PRnge := PRANGE (Buffer);
      PDtValue := PTIMEDATE (DWORDapi(Buffer) + sizeOf(USHORT)*2);
      if PRnge^.ListEntries = 0 then Result := Null
      else if PRnge^.ListEntries = 1 then Result := VarFromDateTime (NotesToDateTime (PDtValue^))
      else begin
        Result := VarArrayCreate ([0, PRnge^.ListEntries-1], varDate);
        for i := 0 to PRnge^.ListEntries-1 do begin
          Result[i] := VarFromDateTime (NotesToDateTime (PDtValue^));
          PDtValue := PTIMEDATE (DWORDapi(PDtValue) + sizeof(TIMEDATE));
        end;
      end;
    end;
    else Result := UNASSIGNED;
  end;
end;

(******************************************************************************)
function UNIDtoStr;
begin
  If fDelimiters
    then Result := intToHex(anUNID.aFile.T2, 8) + ':' +
		               intToHex(anUNID.aFile.T1, 8) + '-' +
                   intToHex(anUNID.Note.T2, 8) + ':' +
                   intToHex(anUNID.Note.T1, 8)
    else Result := intToHex(anUNID.aFile.T2, 8) +
                   intToHex(anUNID.aFile.T1, 8) +
                   intToHex(anUNID.Note.T2, 8) +
                   intToHex(anUNID.Note.T1, 8);
end;

(******************************************************************************)
{ Table conversion }
type
  TConvProc = procedure (Name: AnsiString; Value: variant; Data: pointer);

// This function loops throught the item table and calls Proc for each item found
// passing the item's name, it's value (usually string) and supplied context data
procedure ReadItemTable (pTable: PITEM_TABLE; Proc: TConvProc; Data: pointer);
var
 n: integer;
 ptItem: PITEM;
 pSummary: PAnsiChar;
 name_buf: AnsiString;
 val:   variant;
 wSize: DWORDapi;
begin
 ptItem := PITEM(DWORDapi(pTable) + ODSLength(_ITEM_TABLE));
 wSize := ODSLength(_ITEM_TABLE) + ODSLength(_ITEM) * pTable^.Items;
 pSummary := PAnsiChar(DWORDapi(pTable) + wSize);
 for n := 1 to pTable^.Items do begin
   name_buf := '';
   setLength(name_buf, ptItem^.NameLength + 2);
   System.AnsiStrings.strLCopy(PAnsiChar(name_buf), pSummary, ptItem^.NameLength);
   name_buf[ptItem^.NameLength+1] := #0;
   name_buf := Lmbcs2Native(System.AnsiStrings.StrPas(PAnsiChar(name_buf)));

   val := BufferToValue (PAnsiChar (DWORDapi(pSummary) + ptItem^.NameLength), ptItem^.ValueLength);
   Proc(name_buf, val, Data);
   inc (wSize, ptItem^.NameLength + ptItem^.ValueLength);
   if wSize > pTable^.Length then
     break;
   pSummary := PAnsiChar(DWORDapi(pSummary) + ptItem^.NameLength + ptItem^.ValueLength);
   ptItem := PITEM(DWORDapi(ptItem) + ODSLength(_ITEM));
 end;
end;

//****************************************************
// Item table callback procedure
// Stores all passed items and values to one string in form <Name>=<Val>
procedure ReadItemTableAsStringProc (Name: AnsiString; Value: variant; Data: pointer);
var
 buf: AnsiString;
 i: integer;
begin
 buf := '';
 if (not VarIsNull(Value)) and (not VarIsEmpty(Value)) then try
   if (VarType(Value) and varArray) = 0 then
     buf := AnsiString(VarAsType(Value, varString))
   else begin
     for i := varArrayLowBound(Value,1) to varArrayHighBound(Value,1) do begin
       buf := buf + AnsiString(VarAsType(Value[i], varString)) + ';';
     end;
   end;
 except
   // Ignore conversion errors
   buf := '';
 end;
 PAnsiString(Data)^ := PAnsiString(Data)^ + #13#10 + Name + '=' + buf;
end;

//**********************************************
// Adds an item with unlimited length - by Andy
procedure HugeNSFItemAppend(hNote: NOTEHANDLE;
                          ItemFlags: Word;
                          Name: PAnsiChar;
                          NameLength: Word;
                          DataType: Word;
                          Value: Pointer;
                          ValueLength: LongInt);
const
  MaxBufLen = 65000;
var
  r: longint;
  b: longint;
  l: longint;
  p: pointer;
begin
  if ValueLength <= MaxBufLen then CheckError (NSFItemAppend(hNote, ItemFlags, Name, NameLength,
    DataType, Value, ValueLength))
  else begin
    r := ValueLength;   // the rest of the length
    b := 0;             // the startposition of the pointer in  FContext
    repeat
      if r > MaxBufLen
        then l := MaxBufLen // the length to be inserted now
        else l := r;        // the length to be inserted now
      p:= pointer (longint (Value) + b);
      Inc (b, l);
      CheckError (NSFItemAppend(hNote, ItemFlags, Name, NameLength,
      DataType, p, l ));
      Dec (r, l);
    until (r = 0);
  end;
end;

(******************************************************************************)
{ String conversion }
(******************************************************************************)
function NotesToString(NotesStr: PAnsiChar; StrSize: integer): string;
var
 p:   PAnsiChar;
 s:   string;
 ofs: DWORDapi;
begin
 // Notes uses #0 as CRLF and #0#0 as real line end
 // But it might return normal string as well
 p := NotesStr;
 Result := '';
 ofs := 0;
 if StrSize = 0 then
   exit;

  // Expand
 while True do begin
   s := String(System.AnsiStrings.strPas(p));
   Result := Result + s;
   inc(ofs, length(s) + 1);
   if ofs > DWORDapi(StrSize) then
     break;
   p := PAnsiChar(DWORDapi(NotesStr) + ofs);
   if p^ = #0 then
     break;
   Result := Result + #13#10;
 end;
end;

(******************************************************************************)
function StringToNotes(DelphiStr: string;
               var NotesStr: PAnsiChar;
               var StrSize: integer): boolean;
var
 i, n: integer;
 ansiDelphiStr: AnsiString;
begin
 // Remove LFs leaving CRs
 n := Pos(#10, DelphiStr);
 while n > 0 do begin
   delete(DelphiStr,n,1);
   n := Pos(#10, DelphiStr);
 end;
 ansiDelphiStr := AnsiString(DelphiStr);

 // Copy to buffer and replace CRs with #0
 StrSize := length(ansiDelphiStr) + 2;
 GetMem(NotesStr,StrSize);
 Result := False;
 try
   System.AnsiStrings.strCopy(NotesStr,PAnsiChar(ansiDelphiStr));
   for i := 0 to StrSize-3 do begin
     if NotesStr[i] = #13 then begin
       NotesStr[i] := #0;
       Result := True;
     end;
   end;
   if Result then
     NotesStr[StrSize-1] := #0   //double-zero string
   else begin
     // No internal zeros - normal string
     dec(StrSize);
     ReallocMem(NotesStr,StrSize);
   end;
 except
   FreeMem(NotesStr);
   raise;
 end;
end;

(******************************************************************************)
{ Color conversion }
(******************************************************************************)
const
  ColorMap: array[0..MAX_NOTES_SOLIDCOLORS-1] of integer = (
    $000000, $FFFFFF, $0000FF, $00FF00, $FF0000, $FF00FF, $00FFFF, $FFFF00,
    $000080, $008000, $800000, $800080, $008080, $808000, $808080, $C0C0C0
  );
const
  UNKNOWN_COLOR = $1FFFFFFF;

(******************************************************************************)
function NotesToColor(NotesColor: integer): integer;
begin
  if NotesColor < MAX_NOTES_SOLIDCOLORS
    then Result := ColorMap[NotesColor]
    else Result := UNKNOWN_COLOR; 
end;

(******************************************************************************)
function ColorToNotes(DelphiColor: integer): integer;
begin
  for Result := System.Low(ColorMap) to System.High(ColorMap) do
    if ColorMap[Result] = DelphiColor then exit;
  Result := UNKNOWN_COLOR;
end;

(******************************************************************************)
// TNotesDocumentCollection
(******************************************************************************)
type
  TIDInfo = class
    ID: NOTEID;
    Summary: string;
  end;

(******************************************************************************)
procedure TNotesDocumentCollection.Clear;
var
  i: integer;
begin
  for i := count-1 downto 0 do begin
    TIDInfo(Items[i]).free;
    //FreeMem(PTIdInfo(Items[i]));
    Items[i] := nil;
  end;
  inherited Clear;
end;

(******************************************************************************)
constructor TNotesDocumentCollection.Create;
begin
  inherited Create;
  Db := notesDatabase;
end;

(******************************************************************************)
procedure TNotesDocumentCollection.DeleteDocumentID;
begin
  TIDInfo(Items[Index]).free;
  Delete(Index);
end;

(******************************************************************************)
function TNotesDocumentCollection.GetDocumentId;
begin
  Result := TIdInfo(Items[Idx]).ID;
end;

(******************************************************************************)
function TNotesDocumentCollection.GetDocument;
begin
  Result := TNotesDocument.Create(Db,TIdInfo(Items[Idx]).ID);
  Result.FIsRead := not FUnreadDocs;
  Result.FSummary.Text := TIdInfo(Items[Idx]).Summary;
end;

(******************************************************************************)
function TNotesDocumentCollection.AddDocumentId;
var
  pRec: TIdInfo;
begin
  Result := -1;
  if ID <> 0 then begin
    pRec := TIDInfo.create;
    pRec.ID := ID;
    pRec.Summary := Summary;
    Result := Add (pRec);
  end;
end;

(******************************************************************************)
destructor TNotesDocumentCollection.Destroy;
begin
  Clear;
  inherited Destroy;
end;

//****************************************************
procedure TNotesDocumentCollection.GetCollectionByHandle;
var
  i: integer;
  ptList: PNOTEID;
  ptItems: PITEM_TABLE;
  ID: NOTEID;
  Summary: string;
begin
  PtList := PNOTEID(OSLockObject(hBuffer));
  try
    for i := 1 to NotesFound do begin
      ID := PtList^;
      ptList := PNOTEID(DWORDapi(PtList) + ODSLength(_NOTEID));
      ptItems := PITEM_TABLE(PtList);
      Summary := '';
      if fSummary then begin
        ReadItemTable (ptItems, ReadItemTableAsStringProc, @Summary);
        PtList := PNOTEID(DWORDapi(PtList) + PITEM_TABLE(PtList)^.Length);
      end;
      if ((ID and NOTEID_CATEGORY) = 0) and ((ID and NOTEID_CATEGORY_TOTAL) = 0) then begin
        AddDocumentID (ID, Summary);
      end;
    end;
  finally
    OSUnLockObject(hBuffer);
  end;
end;

(******************************************************************************)
function TNotesDocumentCollection.GetSummaryValues;
begin
  Result := TIdInfo(Items[Idx]).Summary;
end;

//****************************************************
procedure TNotesDocumentCollection.MarkAllRead(fRead: boolean);
begin
  Database.MarkAllRead (Self, fRead);
end;

(******************************************************************************)
// TNotesDatabase
(******************************************************************************)
constructor TNotesDatabase.Create;
begin
  inherited Create;
end;

(******************************************************************************)
destructor TNotesDatabase.Destroy;
begin
  Close;
  inherited Destroy;
end;

//***************************************************
function TNotesDatabase.GetTitle: String;
var
  ansiBuf, ansiResult: AnsiString;
begin
  ansiBuf := '';
  SetLength (ansiBuf, NSF_INFO_SIZE+1);
  CheckError (NSFDbInfoGet (Handle, PAnsiChar(ansiBuf)));
  SetLength (ansiResult, 256);
  NSFDbInfoParse (PAnsiChar(ansiBuf), INFOPARSE_TITLE, PAnsiChar(ansiResult), 255);
  Result := String(System.AnsiStrings.StrPas(PAnsiChar(Lmbcs2Native(ansiResult))));
end;

(******************************************************************************)
function TNotesDatabase.UnreadDocuments: TNotesDocumentCollection;
var
  hTable: LHANDLE;
  id: DWORDapi;
  fFirst: boolean;
  uName: string;
  ansiName: AnsiString;
begin
  // Get IDs
  uName := LmbcsUserName;
  if not TNotesName.IsCanonical (uName) then
    uName := TNotesName.TranslateName (uName, True, '');
  Result := TNotesDocumentCollection.create (self);
  Result.fUnreadDocs := True;
  try
    hTable := 0;
    ansiName := AnsiString(uName);
    CheckError (NSFDbGetUnreadNoteTable(Handle, PAnsiChar(ansiName), length(ansiName), True, @hTable));
    if hTable <> 0 then try
      CheckError (NSFDbUpdateUnread(Handle, hTable));

      // Scan ID table
      id := 0;
      fFirst := True;
      while IDScan (hTable, fFirst, @id) do begin
        fFirst := False;
        Result.AddDocumentId (id, '');
      end;
    finally
      IDDestroyTable(hTable);
    end;
  except
    Result.free;
    raise;
  end;
end;

(******************************************************************************)
procedure TNotesDatabase.UpdateViews;
var
  i: integer;
begin
  if (FViews <> nil) then begin
    for i := 0 to FViews.count-1 do FViews.Objects[i].free;
    FViews.free;
    FViews := nil;
  end;
end;

(******************************************************************************)
class function TNotesDatabase.UserName;
begin
  Result := Lmbcs2Native(LmbcsUserName);
end;

(******************************************************************************)
class function TNotesDatabase.LmbcsUserName: String;
var
 ansiResult: AnsiString;
begin
  setlength(ansiResult,strLength);
  SECKFMGetUserName(PAnsiChar(ansiResult));
  Result := String(System.AnsiStrings.StrPas(PAnsiChar(ansiResult)));
end;

(******************************************************************************)
class function TNotesDatabase.MailFileName: String;
var
 ansiResult: AnsiString;
begin
 setlength(ansiResult,MAXENVVALUE+1);
 OSGetEnvironmentString (MAIL_MAILFILE_ITEM, PAnsiChar(ansiResult), MAXENVVALUE);
 Result := String(System.AnsiStrings.StrPas(PAnsiChar(ansiResult)));
end;

(******************************************************************************)
class function TNotesDatabase.MailServer: String;
var
 ansiResult: AnsiString;
begin
 setlength(ansiResult,MAXENVVALUE+1);
 OSGetEnvironmentString (MAIL_MAILSERVER_ITEM, PAnsiChar(ansiResult), MAXENVVALUE);
 Result := String(System.AnsiStrings.StrPas(PAnsiChar(ansiResult)));
end;

//***************************************************
procedure TNotesDatabase.ListAddressBooks(aServer: string; List: TStrings; fGetTitles: boolean);
var
  flags,wCount,wLength, i: word;
  hReturn: LHANDLE;
  pBuf, pSrv: PAnsiChar;
  buf, title: AnsiString;
begin
  if not fGetTitles then
    flags := 0
  else
    flags := NAME_GET_AB_TITLES or NAME_DEFAULT_TITLES;
  if aServer = '' then
    pSrv := nil
  else
    pSrv := PAnsiChar(AnsiString(aServer));
  CheckError (NAMEGetAddressBooks (pSrv, flags, wCount, wLength, hReturn));
  if hReturn <> 0 then try
    pBuf := OsLockObject(hReturn);
    for i := 1 to wCount do begin
      buf := System.AnsiStrings.strPas(pBuf);
      pBuf := PAnsiChar(DWORDapi(pBuf) + DWORDapi(length(buf) + 1));
      if fGetTitles then begin
        title := AnsiString(System.AnsiStrings.strPas(pBuf));
        buf := buf + '=' + title;
        pBuf := PAnsiChar(DWORDapi(pBuf) + DWORDapi(length(title) + 1));
      end;
      List.add(String(Lmbcs2Native(buf)));
    end;
  finally
    OsUnlockObject (hReturn);
    OsMemFree (hReturn);
  end;
end;

//***************************************************
function TNotesDatabase.Open(aServer, dbFile: string): Boolean;
var
  openingmail:boolean;
  Error: word;
  FilePath: string;
  ansiFilePath, ansiDBFile: AnsiString;
begin
  Close;
  openingmail := false;
  if UpperCase(aServer) = 'LOCAL' then
    aServer := '';
  if (aServer = '') and (dbFile = '') then begin
    aServer := Lmbcs2Native(MailServer);
    dbFile  := Lmbcs2Native(MailFileName);
    openingmail := true;
  end;
  if aServer='' then
    FilePath := dbFile
  else
    FilePath := ConstructPath (TNotesName.TranslateName((aServer), False, ''), dbFile);

  ansiFilePath := Native2Lmbcs(AnsiString(FilePath));
  error := NSFDbOpen(pAnsiChar(ansiFilePath), @FHandle);
  if (error <> 0) and OpeningMail then begin
   { Next line commented out by Noah Silva 10.1.1999     }
   { We don't want the mail.box! we want the MAIL FILE!! }
   {    dbfile := 'mail.box';      }
    aServer:='';
    ansiDBFile := AnsiString(dbfile);
    error := NSFDbOpen(PAnsiChar(ansiDBFile), @FHandle);
  end;
  if ((Handle=0) and (Error=0)) or (Error=USER_CANCEL) then
    raise ELotusNotes.CreateErr(0,'User cancelled Notes session');
  CheckError (Error);

  FFileName   := dbFile;
  FServerName := aServer;
  Result := true;
end;

(******************************************************************************)
procedure TNotesDatabase.Close;
var
 i: integer;
begin
 if not Active then
   exit;
 CheckError(NSFDbClose(Handle));
 FHandle := 0;
 if FViews <> nil then begin
   for i := 0 to FViews.count-1 do
     FViews.Objects[i].free;
   FViews.free;
   FViews := nil;
 end;
end;

(******************************************************************************)
procedure TNotesDatabase.CloseSession;
var
  i: integer;
begin
  if not Active then exit;
  CheckError(NSFDbCloseSession(Handle));
  NSFDbClose(Handle);
  FHandle := 0;
  if FViews <> nil then begin
    for i := 0 to FViews.count-1 do FViews.Objects[i].free;
    FViews.free;
    FViews := nil;
  end;
end;

(******************************************************************************)
procedure TNotesDatabase.OpenMail;
begin
  Open ('','');
end;

(******************************************************************************)
procedure TNotesDatabase.SetActive(Value: boolean);
begin
  if Value <> Active then
    if Value then Open (Server, FileName) else Close;
end;

(******************************************************************************)
function TNotesDatabase.GetACL: TNotesACL;
begin
 if FAcl = nil then
   FAcl := TNotesACL.create(self);
 Result := FACL;
end;

(******************************************************************************)
procedure TNotesDatabase.CopyACL(SourceDb: TNotesDatabase);
begin
  CheckError (NSFDbCopyACL (SourceDb.Handle, Handle));
end;

(******************************************************************************)
function TNotesDatabase.GetActive: Boolean;
begin
  Result := Handle <> 0;
end;

(******************************************************************************)
function SearchProc (Obj: pointer; search_info: pSEARCH_MATCH; summary_info: pITEM_TABLE): STATUS; far; stdcall;
var
  summary: string;
begin
  if ((search_info.SERetFlags and SE_FMATCH) <> 0) then begin
    summary := '';
    if summary_info <> nil then begin
      ReadItemTable(summary_info, ReadItemTableAsStringProc, @summary);
    end;
    TNotesDocumentCollection(Obj).AddDocumentId (search_info^.ID.NoteID, summary);
  end;
  Result := NOERROR;
end;

//***************************************************
function TNotesDatabase.FindDocument;
var
  Coll: TNotesDocumentCollection;
begin
  Coll := Search (Formula, 0, 1);
  try
    if (Coll = nil) or (Coll.Count = 0)
      then Result := nil
      else Result := Coll.Document[0];
  finally
    Coll.free;
  end;
end;

(******************************************************************************)
function TNotesDatabase.FTSearch(query: string;
                maxDocs, sortOptions, otherOptions: integer) :
                TNotesDocumentCollection;
var
  HSearch: THandle;
  retNumDocsFound: DWORDapi;
  retHResults: THandle;
  //hIdTable: THandle;
  pSearchResults: ^FT_SEARCH_RESULTS;
  pNoteId: ^NoteId;
  i: integer;
  ansiQuery: AnsiString;
begin
  Result := TNotesDocumentCollection.Create (self);
  try
    checkError(FTOpenSearch (@HSearch));
    ansiQuery := Native2Lmbcs(AnsiString(Query));
    try
      checkError(Util_LNApi.FTSearch(Handle,
                                    @HSearch,
                                    NullHandle,
                                    PAnsiChar(ansiQuery),
                                    SortOptions or
                                      OtherOptions or
                                      FT_SEARCH_SCORES or
                                      FT_SEARCH_STEM_WORDS,
                                    0,
                                    NullHandle,
                                    @retNumDocsFound,
                                    nil,
                                    @retHResults
                                    ));

      // Well, I don't know why they created an ID table...
      // KOL - 09.03.99
      //checkError(IdCreateTable (sizeOf(NoteId),@hIdTable));
      //try

      pSearchResults := OsLockObject (retHResults);
      try
        pNoteId := pointer(pSearchResults);
        pNoteId := pointer(DWORDapi(pNoteId) + sizeOf (FT_SEARCH_RESULTS));
        for i := 0 to pSearchResults^.NumHits-1 do begin
          Result.AddDocumentId(pNoteID^, '');
          inc(pNoteId);
        end;
      finally
        OsUnlockObject (retHResults);
        OsMemFree (retHResults);
      end;

      {finally
        OsMemFree (hIdTable);
      end;}

    finally
      FTCloseSearch (HSearch);
    end;
  except
    Result.free;
    raise;
  end;
end;

//****************************************************
procedure TNotesDatabase.ReplyMail;
var
  MailDoc: TNotesDocument;
begin
  MailDoc := TNotesDocument.CreateResponse (self, Mail);
  try
    MailDoc.Form := 'Reply';
    MailDoc.SendTo := Mail.MailFrom;
    MailDoc.Subject := 'Re: ' + Mail.Subject;
    MailDoc.BodyAsString := Body;
    MailDoc.Sign;
    MailDoc.Send (False, '');
  finally
    MailDoc.free;
  end;
end;

//***************************************************
function TNotesDatabase.Search;
begin
  Result := FindNotes (formula, notesDateTime, NOTE_CLASS_DATA, True);
end;

//****************************************************
function TNotesDatabase.GetViewByIndex;
begin
  CheckViews;
  Result := TNotesView(FViews.Objects[Index]);
  if Result = nil then begin
    Result := TNotesView.OpenView(self, FViews[Index]);
    FViews.Objects[Index] := Result;
  end;
end;

//****************************************************
procedure TNotesDatabase.CheckViews;
var
  Coll: TNotesDocumentCollection;
  i: integer;
  Doc: TNotesDocument;
begin
  if FViews = nil then begin
    FViews := TStringList.create;
    Coll := FindNotes ('@All', 0, NOTE_CLASS_VIEW or NOTE_CLASS_PRIVATE, False);
    try
      for i := 0 to Coll.count-1 do begin
        Doc := Coll.Document[i];
        try
          FViews.Add (Doc['$TITLE'].AsString);
        finally
          Doc.free;
        end;
      end;
    finally
      Coll.free;
    end;
  end;
end;

//****************************************************
function TNotesDatabase.GetViews (ViewName: string): TNotesView;
// Changed by ap@svd-online.com
const
  AliasSeparator = '|';
var
  i: integer;
  aPos: integer;
begin
  CheckViews;
//  i := FViews.indexOf(ViewName);
//  if i = -1 then Result := nil else Result := ViewByIndex[i];

  Result := nil;
  I      := FViews.IndexOf(ViewName);
  if (I >= 0) then Result := ViewByIndex[I]
  else for I := 0 to FViews.Count - 1 do begin
    aPos := Pos (AliasSeparator, FViews.Strings[I]);
    if (aPos > 0) then begin
      if (0 = SysUtils.CompareText(ViewName,
        trim(Copy(FViews.Strings[I], aPos + 1, System.Length(FViews.Strings[I]) - aPos))))
        then begin
          Result:= ViewByIndex[I];
          Break;
        end;
    end;{of if}
  end;{of for}
end;

//****************************************************
function TNotesDatabase.GetViewCount: Integer;
begin
  CheckViews;
  Result := FViews.count;
end;

//****************************************************
function TNotesDatabase.FindNotes(formula: string;
               notesDateTime: TDateTime;
               noteClass: word;
               fSummary: boolean): TNotesDocumentCollection;
var
  wdc2: status;
  wdc1,wdc3,wdc4,wdc5,wdc6, flags: WORD;
  tdSince: TIMEDATE;
  pSince: PTIMEDATE;
  formula_handle: FORMULAHANDLE;
  ansiFormula: AnsiString;
begin
  Result := TNotesDocumentCollection.Create(self);
  try
    if Formula='' then Formula := '@All';
    if fSummary then flags := SEARCH_SUMMARY else flags := 0;
    if notesDateTime = 0 then pSince := nil
    else begin
      tdSince := DateTimeToNotes(notesDateTime);
      pSince := @tdSince;
    end;

    ansiFormula := Native2Lmbcs(AnsiString(formula));
    try
      CheckError(NSFFormulaCompile(
                 nil,                   { name of formula (none) }
                 WORD(0),               { length of name }
                 PAnsiChar(ansiFormula),    { the ASCII formula }
                 WORD(length(ansiFormula)), { length of ASCII formula }
                 @formula_handle,       { handle to compiled formula }
                 @wdc1,                 { compiled formula length (don't care) }
                 @wdc2,                  { return code from compile (don't care) }
                 @wdc3, @wdc4, @wdc5, @wdc6)); { compile error info (don't care) }

      CheckError(NSFSearch(
                 Handle,         { database handle }
                 formula_handle, { selection formula }
                 nil,            { title of view in selection formula }
                 flags,          { search flags }
                 noteClass,      { note class to find }
                 pSince,         { starting date }
                 SearchProc, { call for each note found }
                 //pointer(self),  { argument to print_fields }
                 pointer(Result),  { argument to print_fields }
                 nil));          { returned ending date (unused) }
    finally
      OSMemFree (formula_handle);
    end;
  except
    Result.free;
    raise;
  end;
end;

//***************************************************
procedure TNotesDatabase.OpenPrivateAddressBook;
var
  BookName: string;
  n: integer;
  ansiVarName, ansiBookName: AnsiString;
begin
  BookName := '';
  ansiVarName  := AnsiString('NAMES');
  SetLength (ansiBookName, 255);
  if OsGetEnvironmentString (PAnsiChar(ansiVarName), PAnsiChar(ansiBookName), 254) then begin
    BookName := String(System.AnsiStrings.strPas(PAnsiChar(ansiBookName)));
    n := Pos (',', BookName);
    if n > 0 then
      System.Delete(BookName, n, length(BookName)-n+1);
  end;
  BookName := Trim(BookName);
  if BookName = '' then
    BookName := 'NAMES.NSF';
  Close;
  Open ('', BookName);
end;

//***************************************************
function TNotesDatabase.OpenView(AName: string; notesDateTime: TDateTime; maxDocs: integer): TNotesView;
begin
  Result := TNotesView.OpenView (self, aName);
end;

//***************************************************
function TNotesDatabase.CreateDocument: TNotesDocument;
begin
  Result := TNotesDocument.CreateNew(self);
end;

//***************************************************
function TNotesDatabase.CreateResponseDocument(const aResUNID: UNID; aSubject: string): TNotesDocument;
begin
  Result := TNotesDocument.CreateResponseByUNID(self, aResUNID);
  Result.Subject := aSubject;
end;

//***************************************************
procedure TNotesDatabase.CreateNew(aServer, dbFile: string; TemplateDB: string);
var
 FilePath: string;
 error: STATUS;
 ansiFilePath, ansiTemplateDB: AnsiString;
begin
 Close;
 if UpperCase(aServer) = 'LOCAL' then
   aServer := '';
 if dbFile = '' then
   raise ELotusNotes.create ('Database Name is blank');
 if aServer = '' then
   FilePath := dbFile
 else
   FilePath := ConstructPath(aServer, dbFile);
 ansiTemplateDB := AnsiString(TemplateDB);
 ansiFilePath   := AnsiString(FilePath);
 if TemplateDB <> '' then begin
   Error := NsfDbCreateAndCopy(PAnsiChar(Native2Lmbcs(ansiTemplateDB)), PAnsichar(ansiFilePath), NOTE_CLASS_ALL, 0, 0, @FHandle);
 end
 else begin
   ansiFilePath := Native2Lmbcs(ansiFilePath);
   CheckError (NsfDbCreate(PAnsiChar(ansiFilePath), DBCLASS_NOTEFILE, True));
   Error := NsfDbOpen (PAnsichar(ansiFilePath), @FHandle);
 end;
 if (Error=USER_CANCEL) or ((Handle=0) and (Error=0)) then
   raise ELotusNotes.CreateErr(0,'User cancelled Notes session');
 CheckError(Error);
 FFileName   := dbFile;
 FServerName := aServer;
end;

//***************************************************
function TNotesDatabase.GetDatabaseID: DBID;
begin
 FillChar(Result, sizeOf(Result), 0);
 CheckError(NsfDBIdGet(Handle, @Result));
end;

//***************************************************
function TNotesDatabase.GetFullName: String;
begin
 if (UpperCase(Server) = 'LOCAL') or (Server = '') then
   Result := FileName
 else
   Result := ConstructPath(Server, FileName);
end;

//***************************************************
function TNotesDatabase.NotesVersion: Word;
begin
 if not Active then
   Open ('', '');
 CheckError (NSFDbGetBuildVersion (Handle, @Result));
end;

//***************************************************
procedure TNotesDatabase.CopyRecords(SourceDB: TNotesDatabase);
var
 Td: TimeDate;
begin
 TimeConstant(TIMEDATE_WILDCARD, Td);
 CheckError (NSFDbCopy (SourceDb.Handle, Handle, TD, NOTE_CLASS_DOCUMENT));
end;

//***************************************************
procedure TNotesDatabase.Delete(DocID: integer);
begin
 CheckError (NSFNoteDelete (Handle, DocId, UPDATE_NOCOMMIT));
end;

//***************************************************
procedure TNotesDatabase.DeleteDocument(Doc: TNotesDocument);
var
 h: integer;
begin
 if Doc.fHandle <> 0 then begin
   h := Doc.DocID;
   NsfNoteClose(Doc.FHandle);
   Doc.fHandle := 0;
   Delete (h);
 end;
end;

//***************************************************
procedure TNotesDatabase.SetTitle(Value: string);
var
 ansiBuf, ansiValue: AnsiString;
begin
 ansiBuf := '';
 SetLength (ansiBuf, NSF_INFO_SIZE+1);
 CheckError (NSFDbInfoGet (Handle, PAnsiChar(ansiBuf)));
 ansiValue := Native2Lmbcs(AnsiString(Value));
 NSFDbInfoModify(PAnsiChar(ansiBuf), INFOPARSE_TITLE, PAnsiChar(ansiValue));
 CheckError( NSFDbInfoSet(Handle, PAnsiChar(ansiBuf)));
end;

//***************************************************
procedure TNotesDatabase.SetFileName(Value: string);
begin
 if Value <> FileName then begin
   Close;
   FFileName := Value;
 end;
end;

//***************************************************
procedure TNotesDatabase.MarkRead(NoteID: DWORDapi; fRead: boolean);
var
 uName: string;
 hTable, hOriginalTable: THandle;
 ansiName: AnsiString;
begin
 // Get IDs
 hOriginalTable := 0;
 uName := self.LmbcsUserName();
 if not TNotesName.IsCanonical(uName) then
   uName := TNotesName.TranslateName (uName, True, '');
 ansiName := AnsiString(uName);
 CheckError (NSFDbGetUnreadNoteTable(Handle, PAnsiChar(ansiName), length(ansiName), True, @hTable));
  try
    // Bring table up to date
    CheckError (NSFDbUpdateUnread (Handle, hTable));

    // Notes requires original unread table to merge changes
    CheckError (IDTableCopy (hTable, @hOriginalTable));

    // Delete or insert ID
    if fRead
      then IDDelete (hTable, NoteID, nil)
      else IDInsert (hTable, NoteID, nil);

    // Merge and update table
    CheckError (NSFDbSetUnreadNoteTable (Handle, PAnsiChar(ansiName), length(ansiName),
                      False, hOriginalTable, hTable));
  finally
    if hTable <> 0 then IDDestroyTable(hTable);
    if hOriginalTable <> 0 then IDDestroyTable(hOriginalTable);
  end;
end;

//***************************************************
procedure TNotesDatabase.MarkAllRead(Docs: TNotesDocumentCollection; fRead: boolean);
var
 uName: String;
 ansiName: AnsiString;
 hTable, hOriginalTable: THandle;
 i: integer;
begin
 // Get IDs
 hOriginalTable := 0;
 uName := LmbcsUserName;
 if not TNotesName.IsCanonical (uName) then
   uName := TNotesName.TranslateName (uName, True, '');
 ansiName := AnsiString(uName);
 CheckError( NSFDbGetUnreadNoteTable(Handle, PAnsichar(ansiName), Length(ansiName), True, @hTable) );
 try
   // Bring table up to date
   CheckError( NSFDbUpdateUnread(Handle, hTable) );
   // Notes requires original unread table to merge changes
   CheckError( IDTableCopy(hTable, @hOriginalTable) );
   // Delete or insert IDs
   for i := 0 to Docs.count-1 do begin
     if fRead then
       IDDelete (hTable, Docs.DocumentId[i], nil)
     else
       IDInsert (hTable, Docs.DocumentId[i], nil);
   end;
   // Merge and update table
   CheckError( NSFDbSetUnreadNoteTable(Handle, PAnsichar(ansiName), length(ansiName),
                                       False, hOriginalTable, hTable));
 finally
   if hTable <> 0 then
     IDDestroyTable(hTable);
   if hOriginalTable <> 0 then
     IDDestroyTable(hOriginalTable);
 end;
end;

//***************************************************
procedure TNotesDatabase.SetReplicaInfo;
begin
  CheckError(NsfDbReplicaInfoSet(Handle, @Value));
end;

//***************************************************
function TNotesDatabase.GetReplicaInfo;
begin
  FillChar(Result, sizeOf(Result), 0);
  CheckError(NsfDbReplicaInfoGet(Handle, @Result));
end;

//***************************************************
procedure TNotesDatabase.SetServerName(Value: string);
begin
  if Value <> Server then begin
    Close;
    FServerName := Value;
  end;
end;

//**********************************************
procedure TNotesDatabase.SendMail;
var
  MailDoc: TNotesDocument;
begin
  MailDoc := CreateDocument;
  try
    MailDoc.Form := 'Memo';
    MailDoc.SendTo := Address;
    MailDoc.Subject := Subject;
    MailDoc.BodyAsString := Body;
    MailDoc.Sign;
    MailDoc.Send (False, '');
  finally
    MailDoc.free;
  end;
end;

//***************************************************
function TNotesDatabase.GetQuotaInfo: DBQUOTAINFO;
begin
 FillChar(Result, sizeOf(Result), 0);
 CheckError( NSFDbQuotaGet(PAnsiChar(AnsiString(ConstructPath(Server,FileName))),
               @Result) );
end;
//ClassMarker_Method(TNotesDatabase)

(******************************************************************************)
{TNotesItem}
(******************************************************************************)
constructor TNotesItem.Create(notesDocument: TNotesDocument; aName: string);
begin
  inherited Create;
  fDoc := notesDocument;
  fName:= aName;
  fSaveToDisk := True;
  SetCreated;
end;

(******************************************************************************)
constructor TNotesItem.CreateNext;
begin
  inherited Create;
  fDoc := notesItem.Document;
  fName := notesItem.Name;
  fSaveToDisk := True;
  fCreated := True;
  InitItemInfo(notesItem.ItemBid);
  FSeqNo := notesItem.SeqNo + 1;
end;

(******************************************************************************)
destructor TNotesItem.Destroy;
begin
  FStringsValue.free;
  inherited destroy;
end;

//****************************************************
procedure TNotesItem.GetValueBuffer;
var
  Ptr: pointer;
begin
  Ptr := OSLockBlock (fValueBid);
  try
    GetMem (Buffer, fValueLength);
    Move (Ptr^, Buffer^, fValueLength);
    BufSize := fValueLength;
  finally
    OSUnlockBlock (fValueBid);
  end;
end;

(******************************************************************************)
procedure TNotesItem.InitItemInfo(PrevItem: BLOCKID);
var
 ansiRetName: AnsiString;
 retNameLength: word;
 ansiName2: AnsiString;
begin
 ansiName2 := Native2Lmbcs(AnsiString(Name));
 if (PrevItem.pool = 0) and (PrevItem.block = 0) then begin
   // First item in sequence
   CheckError( NSFItemInfo (Document.Handle,
                             PAnsiChar(ansiName2),
                             Length(ansiName2),
                             @fItemBId,
                             @fDataType,
                             nil,
                             nil
                             ) );
 end
 else begin
   // Next item
   CheckError( NSFItemInfoNext(Document.Handle,
                             PrevItem,
                             PAnsiChar(ansiName2),
                             Length(ansiName2),
                             @fItemBId,
                             @fDataType,
                             nil,
                             nil
                             ) );
 end;
 SetLength(ansiRetName,256);
 NSFItemQuery( Document.Handle,
                fItemBID,
                PAnsiChar(ansiRetName),
                255,
                @retNameLength,
                @fItemFlags,
                @fDataType,
                @fValueBID,
                @fValueLength
 );
end;

//**********************************************
procedure TNotesItem.SetAsTimes;
var
  i: integer;
  v: variant;
begin
  if (VarType(Value) and varArray) = 0 then exit;
  if (VarType(Value) and varDate) <> 0 then AsList := Value
  else begin
    v := VarArrayCreate ([VarArrayLowBound(Value,1), VarArrayHighBound(Value,1)], varDate);
    for i := VarArrayLowBound(Value,1) to VarArrayHighBound(Value,1) do
      v[i] := VarAsType (Value[i], varDate);
    AsList := v;
  end;
  SetCreated;
end;

(******************************************************************************)
procedure TNotesItem.SetItemFlags (anItemFlag: integer; anValue: boolean);
begin
  if anValue then fItemFlags := fItemflags or anItemFlag;
end;

//**********************************************
function TNotesItem.GetRichText;
begin
  if FStringsValue = nil then FStringsValue := TStringList.create;
  if fCreated then FStringsValue.Text := AsString;
  Result := FStringsValue;
end;

//**********************************************
function TNotesItem.GetAsTimes;
begin
  if ItemType = TYPE_TIME_RANGE then Result := AsList else Result := NULL;
end;

(******************************************************************************)
function TNotesItem.GetItemFlags (anItemFlag: integer): boolean;
begin
  result := anItemFlag = (fItemFlags and anItemFlag);
end;

(******************************************************************************)
function TNotesItem.GetLastModifed: TDateTime;
var
 TM: TIMEDATE;
 ansiName2: AnsiString;
begin
 Result := 0;
 if fIsNew then
   exit;
 ansiName2 := Native2Lmbcs(AnsiString(Name));
 CheckError( NsfItemGetModifiedTime(Document.Handle,
                                     PAnsiChar(ansiName2),
                                     Length(ansiName2),
                                     0,
                                     @TM
                                     ));
 Result := NotesToDateTime(TM);
end;

(******************************************************************************)

function TNotesItem.GetAsDateTime: TDateTime;
var
 TM: TIMEDATE;
begin
 Result := 0;
 if not fCreated then
   exit;
 if NSFItemGetTime(Document.Handle,PAnsiChar(Native2Lmbcs(AnsiString(Name))), @TM) then
   Result := NotesToDateTime(TM);
end;

//******************************************************************************

function TNotesItem.GetAsFloat: Extended;
var
 ldb: double;
begin
 Result := 0;
 if not fCreated then
   exit;
 if NSFItemGetNumber(Document.Handle,PAnsiChar(Native2Lmbcs(AnsiString(Name))),@ldb) then
   Result := ldb;
end;

//**********************************************
function TNotesItem.GetAsList: Variant;
var
 Ptr: pointer;
 PRnge: PRANGE;
 PNumValue: PNUMBER;
 PDtValue: PTIMEDATE;
 i, rCount: integer;
 ansiName2, ansiStr:  AnsiString;
begin
 Result := NULL;
 if not fCreated then
   exit;
 ansiName2 := Native2Lmbcs(AnsiString(Name));
 case ItemType of
   TYPE_TEXT_LIST: begin
     rCount := NSFItemGetTextListEntries(Document.Handle,PAnsiChar(ansiName2));
     if rcount > 0 then begin
       Result := VarArrayCreate ([0, rCount-1], varOleStr);
       for i := 0 to rCount-1 do begin
         SetLength (ansiStr,2024);
         NSFItemGetTextListEntry (Document.Handle,PAnsiChar(ansiName2),i, PAnsiChar(ansiStr),2023);
         Result[i] := String(Lmbcs2Native(System.AnsiStrings.StrPas(PAnsiChar(ansiStr))));
       end;
     end;
   end;
   TYPE_NUMBER_RANGE: begin
     Ptr := OSLockBlock (fValueBid);
     try
       PRnge := PRANGE (DWORDapi(Ptr) + sizeof(WORD));
       if PRnge^.ListEntries > 0 then begin
         Result := VarArrayCreate ([0, PRnge^.ListEntries-1], varDouble);
         PNumValue := PNUMBER (DWORDapi(PRnge) + sizeOf(USHORT)*2);
         for i := 0 to PRnge^.ListEntries-1 do begin
           Result[i] := PNumValue^;
           PNumValue := PNUMBER (DWORDapi(PNumValue) + sizeof(NUMBER));
         end;
       end;
     finally
       OSUnlockBlock (fValueBid);
     end;
   end;
   TYPE_TIME_RANGE: begin
     Ptr := OSLockBlock (fValueBid);
     try
       PRnge := PRANGE (DWORDapi(Ptr) + sizeof(WORD));
       if PRnge^.ListEntries > 0 then begin
         Result := VarArrayCreate ([0, PRnge^.ListEntries-1], varDate);
         PDtValue := PTIMEDATE (DWORDapi(PRnge) + sizeOf(USHORT)*2);
         for i := 0 to PRnge^.ListEntries-1 do begin
           Result[i] := VarFromDateTime (NotesToDateTime (PDtValue^));
           PDtValue := PTIMEDATE (DWORDapi(PDtValue) + sizeof(TIMEDATE));
         end;
       end;
     finally
       OSUnlockBlock (fValueBid);
     end;
   end;
 end;
end;

//**********************************************
function TNotesItem.GetAsNumbers: Variant;
begin
 Result := 0;
 if not fCreated then
   exit;
 if ItemType = TYPE_NUMBER_RANGE then
   Result := GetAsList;
end;

//****************************************************
function TNotesItem.GetAsReference: UNID;
var
 Buf: pLIST;
 sz: DWORDapi;
begin
 FillChar (Result, sizeOf(Result), 0);
 if not fCreated then
   exit;
 GetValueBuffer (pointer(Buf), sz);
 if Buf = nil then
   exit;

 // Buffer points to list of references now - but I never saw more than 1 ref
 if Buf^.ListEntries >= 1 then
   // CopyMemory (@Result, pointer(DWORDapi(Buf) + ODSLength(_LIST)), sizeOf(UNID));
	 CopyMemory(@Result, pointer(DWORDapi(Buf) + ODSLength(_LIST) * 2), sizeOf(UNID));
end;

//**********************************************
function TNotesItem.GetAsString: String;
var
 hBuffer: LHandle;
 retLen: DWORDapi;
 pBuf:   PAnsiChar;
 ansiName, ansiRet: AnsiString;
begin
 Result := '';
 if not fCreated then
   exit;
 ansiName := AnsiString(Name);
 if ItemType = TYPE_TEXT then begin
   setLength(ansiRet,ValueLength+1);
   retLen := NSFItemGetText (Document.Handle,PAnsiChar(Native2Lmbcs(ansiName)),PAnsiChar(ansiRet),ValueLength);
   Result := NotesToString(PAnsiChar(ansiRet),retLen);
 end
 else if ItemType = TYPE_COMPOSITE then begin
   // by Andy
   CheckError (ConvertItemToText (fValueBID,ValueLength,#13#10,65535, @hBuffer, @retLen, false));
   if hBuffer <> 0 then try
     pBuf:= OsLockObject(hBuffer);
     Result:= String(System.AnsiStrings.StrPas(pBuf));
   finally
     OsUnlockObject(hBuffer);
     OsMemFree (hBuffer);
   end
 end
 else begin
   SetLength(ansiRet,60000);
   NSFItemConvertToText(Document.Handle,
                        PAnsiChar(Native2Lmbcs(ansiName)),
                        PAnsiChar(ansiRet),
                        60000,
                        ITEM_VALUE_SEPARATOR
   );
   ansiRet := System.AnsiStrings.StrPas(PAnsiChar(ansiRet));
 end;
 Result := String(Lmbcs2Native(ansiRet));
end;

//**********************************************
procedure TNotesItem.SetAsFloat(Value: extended);
var
 v: NUMBER;
 ansiName: AnsiString;
begin
 v := Value;
 IsSummary := True;
 ansiName := AnsiString(Name);
 CheckError( NSFItemSetNumber(Document.Handle,PAnsiChar(Native2Lmbcs(ansiName)),@v) );
 SetCreated;
end;

//**********************************************
procedure TNotesItem.SetAsList(Value: variant);
var
 MemPtr: PRANGE;
 ValPtr: PNUMBER;
 DtPtr: PTIMEDATE;
 ArrSz, ValSz, ListSz: DWORDapi;
 i, Lob, Hib: integer;
 ansiName2, ansiStr: AnsiString;
 hList: LHANDLE;
 ListPtr: pointer;
begin
 if (VarType(Value) and varArray) = 0 then
   SetItemValue(Value)
 else begin
   ansiName2 := Native2Lmbcs(AnsiString(Name));
   NsfItemDelete(Document.Handle, PAnsiChar(ansiName2), length(ansiName2));
   HiB := VarArrayHighBound (Value, 1);
   LoB := VarArrayLowBound (Value, 1);
   ArrSz := HiB-LoB + 1;
   case (VarType(Value) and varTypeMask) of
     varString, varOleStr: begin
       ListSz := 0;
       CheckError(ListAllocate(0, 0, False, @hList, @ListPtr, @ListSz));
       try
         OsUnlockObject(hList);
         for i := LoB to HiB do begin
           ansiStr := Native2Lmbcs(AnsiString(Value[i]));
           CheckError(ListAddEntry(hList,False,@ListSz,i,PAnsiChar(ansiStr),length(ansiStr)));
         end;
         ListPtr := OsLockObject(hList);
         CheckError(NSFItemAppend(Document.Handle,ItemFlags,
                        PAnsiChar(ansiName2),length(ansiName2),TYPE_TEXT_LIST,
                        ListPtr,ListSz));
       finally
         OSUnlockObject(hList);
         OSMemFree(hList);
       end;
       {CheckError (NSFItemCreateTextList (Document.Handle,PAnsiChar(Name2), PAnsiChar(str), length(str)));
       str := Native2Lmbcs(Value[LoB]);
       for i := Lob+1 to Hib do begin
         str := Native2Lmbcs(Value[i]);
         CheckError (NSFItemAppendTextList (Document.Handle,PAnsiChar(Name2), PAnsiChar(str),MAXWORD,True));
       end;}
     end;
     varSmallint, varInteger, varSingle, varDouble, varCurrency, varBoolean: begin
        ValSz := 2*sizeof(USHORT) + (ArrSz * sizeOf(NUMBER));
        GetMem (MemPtr, ValSz);
        try
          MemPtr^.ListEntries := ArrSz;
          MemPtr^.RangeEntries := 0;
          ValPtr := PNUMBER (DWORDapi(MemPtr) + 2*sizeOf(USHORT));
          for i := Lob to Hib do begin
            ValPtr^ := NUMBER (Value[i]);
            ValPtr := PNUMBER (DWORDapi(ValPtr) + sizeOf(NUMBER));
          end;
          CheckError (NSFItemAppend (Document.Handle, ITEM_SUMMARY, PAnsiChar(ansiName2), length(ansiName2),
            TYPE_NUMBER_RANGE, MemPtr, ValSz));
        finally
          FreeMem (MemPtr);
        end;
      end;
      varDate: begin
        ValSz := 2*sizeof(USHORT) + (ArrSz * sizeOf(TIMEDATE));
        GetMem (MemPtr, ValSz);
        try
          MemPtr^.ListEntries := ArrSz;
          MemPtr^.RangeEntries := 0;
          DtPtr := PTIMEDATE (DWORDapi(MemPtr) + 2*sizeOf(USHORT));
          for i := Lob to Hib do begin
            DtPtr^ := DateTimeToNotes(VarToDateTime (Value[i]));
            DtPtr := PTIMEDATE (DWORDapi(DtPtr) + sizeOf(TIMEDATE));
          end;
          CheckError (NSFItemAppend(Document.Handle, ITEM_SUMMARY, PAnsiChar(ansiName2), length(ansiName2),
                         TYPE_TIME_RANGE, MemPtr, ValSz));
        finally
          FreeMem (MemPtr);
        end;
      end;
    end;
  end;
  SetCreated;
end;

//**********************************************
procedure TNotesItem.SetAsNumbers(Value: variant);
var
 i: integer;
 v: variant;
begin
 if (VarType(Value) and varArray) = 0 then
   exit;
 if (VarType(Value) and varDouble) <> 0 then
   AsList := Value
 else begin
   v := VarArrayCreate ([VarArrayLowBound(Value,1), VarArrayHighBound(Value,1)], varDouble);
   for i := VarArrayLowBound(Value,1) to VarArrayHighBound(Value,1) do
     v[i] := VarAsType (Value[i], varDouble);
   AsList := v;
 end;
 SetCreated;
end;

//****************************************************
procedure TNotesItem.SetAsReference(Value: UNID);
var
 sz: DWORDapi;
 buf: pLIST;
begin
 sz := ODSLength(_LIST) + sizeOf(UNID);
 GetMem (Buf, sz);
 try
   Buf^.ListEntries := 1;
   CopyMemory (pointer(DWORDapi(Buf) + ODSLength(_LIST)), @Value, sizeOf(Value));
   SetValueBuffer (TYPE_NOTEREF_LIST, ItemFlags, Buf, sz);
 finally
   FreeMem(buf);
 end;
end;

//**********************************************
procedure TNotesItem.SetAsString(Value: string);
var
 buf: PAnsiChar;
 bufsz: integer;
begin
 Value := String(Native2Lmbcs(AnsiString(Value)));
 StringToNotes(Value, buf, bufsz);
 try
   dec(bufsz);   //trailing zero is not counted in
   SetValueBuffer(TYPE_TEXT, ItemFlags, buf, bufsz);
   //SetValueBuffer(TYPE_TEXT, ItemFlags, PAnsiChar(Value), length(Value));
   //CheckError(NSFItemSetText(Document.Handle,PAnsiChar(Native2Lmbcs(Name)),PAnsiChar(Value),length(Value)));
   SetCreated;
 finally
   FreeMem(buf);
 end;
end;

(******************************************************************************)
procedure TNotesItem.SetAsDateTime(Value: TDateTime);
var
 T: TIMEDATE;
 ansiName: AnsiString;
begin
 if value = 0 then
   AsString := ''
 else begin
   T := DateTimeToNotes(Value);
   ansiName := AnsiString(Name);
   CheckError( NSFItemSetTime(Document.Handle, PAnsiChar(Native2Lmbcs(ansiName)), @T) );
   SetCreated;
 end;
end;

(******************************************************************************)
function TNotesItem.GetItemValue: Variant;
var
 ansiStr: AnsiString;
begin
 Result := NULL;
 if not fCreated then
   exit;
 case ItemType of
   TYPE_TEXT:          Result := AsString;
   TYPE_TEXT_LIST:     Result := AsList;
   TYPE_NUMBER:        Result := AsNumber;
   TYPE_NUMBER_RANGE:  Result := AsNumbers;
   TYPE_TIME:          Result := VarFromDateTime (AsDateTime);
   TYPE_TIME_RANGE:    Result := AsTimes;
   TYPE_NOTEREF_LIST:  Result := UNIDToStr(AsReference,True);
   else begin
     SetLength (ansiStr, 62001);
     NSFItemConvertValueToText (ItemType, fValueBid, ValueLength, PAnsiChar(ansiStr), 62000, #0);
     Result := String(Lmbcs2Native(System.AnsiStrings.strPas(PAnsiChar(ansiStr))));
   end;
 end;
end;

(******************************************************************************)
procedure TNotesItem.SetItemValue(aValue: Variant);
begin
 if (VarType(aValue) and varArray) <> 0 then
   AsList := aValue
 else begin
   case (VarType(aValue) and varTypeMask) of
     varEmpty, varNull:   ;
     varSmallint, varInteger:
       AsNumber := int(integer(aValue));
     varSingle, varDouble, varCurrency:
       AsNumber := extended(aValue);
     varDate:
       AsDateTime := VarToDateTime (aValue);
     varBoolean:
       AsNumber := int (ord(boolean (aValue)));
     varString, varOleStr:
       AsString := aValue;
   end;
 end;
 SetCreated;
end;

//**********************************************
constructor TNotesItem.CreateNew;
begin
  inherited Create;
  fDoc := notesDocument;
  fName := aName;
  fSaveToDisk := true;
  IsSummary := True;
  fCreated := False;
end;

//**********************************************
procedure TNotesItem.SetCreated;
begin
 if not fCreated then begin
   InitItemInfo(NullBid);
   fCreated := True;
 end;
end;

//**********************************************
function TNotesItem.GetAsStrings;
var
 Val: variant;
 i: integer;
begin
 if FStringsValue = nil then
   FStringsValue := TStringList.create;
 if fCreated and ((ItemType = TYPE_TEXT_LIST) or
   (ItemType = TYPE_NUMBER_RANGE) or
   (ItemType = TYPE_TIME_RANGE)) then begin
   Val := AsList;
   for i := VarArrayLowBound (Val,1) to VarArrayHighBound(Val,1) do
     FStringsValue.add (Lmbcs2Native(string (Val[i])));
 end;
 Result := FStringsValue;
end;

//**********************************************
procedure TNotesItem.SetAsStrings;
var
  Val: variant;
  i: integer;
begin
  if Value = nil then exit;
  if FStringsValue = nil then FStringsValue := TStringList.create;
  FStringsValue.Assign (Value);
  if FStringsValue.count = 0 then AsString := ''
  else begin
    Val := VarArrayCreate ([0, FStringsValue.count-1], varOLEStr);
    for i := 0 to FStringsValue.count-1 do Val[i] := FStringsValue[i];
    AsList := Val;
  end;
end;

//**********************************************
procedure TNotesItem.SetRichText(Value: TStrings);
var
 cHandle: THandle;
 ansiName2, ansiBuf: AnsiString;
begin
 ansiBuf   := Native2Lmbcs(AnsiString(Value.text));
 ansiName2 := Native2Lmbcs(AnsiString(Name));
 NsfItemDelete(Document.Handle, PAnsiChar(ansiName2), length(ansiName2));
 CheckError (CompoundTextCreate(Document.Handle, PAnsiChar(ansiName2), @cHandle));
 try
   CheckError( CompoundTextAddText(cHandle, STYLE_ID_SAMEASPREV, Default_Font_ID,
                   PAnsiChar(ansiBuf), length(ansiBuf), #13#10,
                   COMP_PRESERVE_LINES, NULLHANDLE));
 finally
   CompoundTextClose (cHandle, nil, nil, nil, 0);
 end;
end;

//****************************************************
procedure TNotesItem.SetValueBuffer(iType,wFlags: word; Buffer: pointer;  BufSize: DWORDapi);
var
 ansiName2: AnsiString;
begin
 ansiName2 := Native2Lmbcs(AnsiString(Name));
 NsfItemDelete(Document.Handle, PAnsiChar(ansiName2), length(ansiName2));
 HugeNsfItemAppend(Document.Handle,wFlags,PAnsiChar(ansiName2),length(ansiName2),
         iType,Buffer,BufSize);
end;

(******************************************************************************)
function TNotesItem.CreateNextItem: TNotesItem;
var
 cls: TNotesItemClass;
begin
 if not NextItemExists then
   Result := nil
 else begin
   cls := TNotesItemClass(self.classType);
   Result := cls.createNext(self);
 end;
end;

(******************************************************************************)
function TNotesItem.NextItemExists: boolean;
var
 ansiName2: AnsiString;
 tDataType: WORD;
 tItemBid:  BLOCKID;
begin
 ansiName2 := Native2Lmbcs(AnsiString(Name));
 Result := NSFItemInfoNext(Document.Handle,
                             ItemBid,
                             PAnsiChar(ansiName2),
                             Length(ansiName2),
                             @tItemBid,
                             @tDataType,
                             nil,
                             nil
                             ) = 0;
end;

(******************************************************************************)

function TNotesItem.LoadNextItem: boolean;
begin
 Result := NextItemExists;
 if Result then begin
   // Trying to load next item info
   InitItemInfo(ItemBid);
   inc(FSeqNo);
 end;
end;

//ClassMarker_Method(TNotesItem)

(******************************************************************************)
{ TNotesDocument}
(******************************************************************************)

procedure TNotesDocument.AttachForm;
var
  Forms: TNotesDocumentCollection;
  FormDoc: TNotesDocument;
  i: integer;
  itname: string;
begin
  // Find form note document
  FormDoc := nil;
  Forms := Database.FindNotes('$TITLE="' + aForm + '"', 0, NOTE_CLASS_FORM, False);
  if Forms.count = 0 then
    raise ELotusNotes.CreateErr(0, 'Form ' + aForm + ' was not found');
  try
    // Copy form items to this document - $Body, $Info, $Title, $Script...
    FormDoc := Forms.Document[0];
    for i := 0 to FormDoc.FieldCount-1 do begin
      itname := upperCase(FormDoc.FieldName[i]);
      //if (itname = '$BODY') or (itname = '$INFO') or (itname = '$TITLE') or
      //(pos (itname, '$$Script') = 1)
      if (itname <> '') and (itname[1] = '$') and (itname <> '$FIELDS') and
      (itname <> '$SIGNATURE') and (itname <> '$UPDATEDBY') and (itname <> '$FLAGS') and
      (itname <> '$REVISIONS') and (pos('COPYTO',itname) = 0) and (pos('SENDTO',itname) = 0)
      then begin
        CopyItem (FormDoc, FormDoc.FieldName[i]);
      end;
    end;
    if IsItemExists('Form') then deleteItem('Form');
  finally
    FormDoc.free;
    Forms.free;
  end;
end;

(******************************************************************************)
function TNotesDocument.ComputeWithForm;
var res : STATUS;
begin
  res := NSFNoteComputeWithForm(Handle,0,0,nil,nil);
  if raiseerror then CheckError(res);
  result := (res = 0)
end;

(******************************************************************************)
constructor TNotesDocument.CreateByUNID;
var
  id: DWORDapi;
begin
  inherited Create;
  CheckError(NSFNoteOpenByUNID(notesDatabase.Handle, @anUNID, 0, @FHandle));
  NSFNoteGetInfo(FHandle, _NOTE_ID, @id);
  InitDocument (NotesDatabase, id);
end;

//****************************************************
constructor TNotesDocument.CreateNew(notesDatabase: TNotesDatabase);
var
 tdDate: TIMEDATE;
begin
 inherited Create;
 FFields := TStringList.create;
 FAttach := TStringList.Create;
 CheckError( NsfNoteCreate(NotesDatabase.Handle, @FHandle) );

 // Set "ComposedDate" to the current time/date right now
 OSCurrentTIMEDATE(@tdDate);
 CheckError (MailAddHeaderItem(Handle, MAIL_COMPOSEDDATE_ITEM_NUM,
             PAnsiChar(@tdDate), WORD(sizeof(TIMEDATE))));
 FDatabase := notesDatabase;
end;

(******************************************************************************)
procedure TNotesDocument.CopyItem;
var
  itSource, itTarget: TNotesItem;
  buf, bufptr: pointer;
  bufsz: DWORDapi;
begin
  if IsItemExists(itemName) then DeleteItem(itemName);
  itSource := Source.Items[itemName];
  if itSource.ItemType = TYPE_COMPOSITE
    then itTarget := TNotesRichTextItem.CreateNew(self, itemName)
    else itTarget := TNotesItem.CreateNew (self, itemName);
  itSource.GetValueBuffer(buf, bufsz);
  if buf <> nil then try
    bufptr := pointer(DWORDapi(buf) + sizeof(WORD));
    itTarget.SetValueBuffer(itSource.ItemType, itSource.ItemFlags, bufptr, bufsz - sizeOf(WORD));
  finally
    FreeMem(buf);
  end;
end;

(******************************************************************************)
function TNotesDocument.CopyToDatabase;
var
  Ddid:NoteId;
begin
  CheckError (NSFDbCopyNote(Database.Handle,nil,nil,DocId,DestDB.Handle,nil,nil,@Ddid,nil));
  Result := TNotesDocument.Create(DestDB,ddid);
end;

(******************************************************************************)
procedure TNotesDocument.CheckAddress;

  procedure DoCheck (Address: String);
  var
    n: integer;
    s: string;
  begin
    while Address <> '' do begin
      n := Pos (',', Address);
      if n = 0 then
        n := length(Address) + 1;
      s := String(Native2Lmbcs(AnsiString(trim(Copy(Address, 1, n-1)))));
      delete (Address, 1, n);

      if not TNotesName.CheckAddress (Database.MailServer, s) then
        raise ELotusNotes.createErr (0, 'Cannot find address "' + s + '" on server ' + Database.MailServer);
    end;
  end;

begin
  DoCheck (SendTo);
  DoCheck (Recipients);
end;

(******************************************************************************)
constructor TNotesDocument.Create;
begin
  inherited Create;
  CheckError(NSFNoteOpen(notesDatabase.Handle, anId, 0, @FHandle));
  InitDocument (NotesDatabase, anID);
end;

(******************************************************************************)
destructor TNotesDocument.Destroy;
var
  i: integer;
begin
  if FHandle <> 0 then NsfNoteClose (FHandle);
  if FAttach <> nil then
    for i := 0 to FAttach.count-1 do
      FAttach.Objects[i].free;
  if FFields <> nil then
    for i := 0 to FFields.count-1 do
      FFields.Objects[i].Free;
  FAttach.free;
  FFields.free;
  FFontTable.free;
  FSummary.free;
  inherited Destroy;
end;

(******************************************************************************)
function TNotesDocument.GetBodyAsString;
begin
  Result := Items['Body'].AsString;
end;

(******************************************************************************)
procedure TNotesDocument.SetBodyAsString(Value: string);
var
 dwBodyItemLen: DWORDapi;
 hBodyItem: integer;
 i:Integer;
 TempList: TStringList;
 ansiStr:  AnsiString;
begin
 if Value <> '' then begin
    //Value := Native2Lmbcs(Value);
    // MailAddBodyItem uses this function automatically - Fujio Kurose
    CheckError (MailCreateBodyItem(@hBodyItem, @dwBodyItemLen));
    TempList := TStringList.Create;
    TempList.Text := Value;
    try
      for i:=0 to TempList.Count-1 do begin
        ansiStr := AnsiString(TempList[i]);
        CheckError (MailAppendBodyItemLine(hBodyItem, @dwBodyItemLen,
                        PAnsiChar(ansiStr), word(length(ansiStr))));
      end;
      CheckError (MailAddBodyItem(Handle, hBodyItem, dwBodyItemLen, nil));
    finally
      OsMemFree(hBodyItem);
      TempList.Free;
    end;
  end;
end;

(******************************************************************************)
function TNotesDocument.GetBodyAsMemo;
begin
  Result := Items['Body'].AsStrings;
end;

(******************************************************************************)
procedure TNotesDocument.SetBodyAsMemo(Value: TStrings);
var
 wBodyCount: word;
 hBodyItem: integer;
 dwBodyItemLen: DWORDapi;
 itemName: string;
 ansiStr: AnsiString;
begin
 itemName := 'Body';
 if Value.Count > 0 then begin
   CheckError (MailCreateBodyItem (@hBodyItem, @dwBodyItemLen));
   try
     for wBodyCount := 0 to Value.Count-1 do begin
       ansiStr := AnsiString(Value[wBodyCount]);
       CheckError( MailAppendBodyItemLine(hBodyItem,@dwBodyItemLen,
                       PAnsiChar(ansiStr),length(ansiStr)));
     end;
     CheckError( MailAddBodyItem(Handle,hBodyItem,dwBodyItemLen, nil));
   finally
     OsMemFree(hBodyItem);
   end;
 end;
end;

(******************************************************************************)
function TNotesDocument.GetItemByNum(ItemNum: integer): string;
var
 error: integer;
 fldLength: word;
 ansiRet: AnsiString;
begin
 ansiRet := '';
 setLength(ansiRet,strLength);
 error := MailGetMessageItem (Handle, ItemNum, PAnsiChar(ansiRet), strLength, @fldLength);
 ansiRet := System.AnsiStrings.strPas(PAnsiChar(ansiRet));
 if error <> 0 then
   Result := ''
 else
   Result := String(Lmbcs2Native(ansiRet));
 if error <> 546 then
   CheckError(error);   //546 - item not found
end;

(******************************************************************************)
procedure TNotesDocument.SetIsRead(Value: boolean);
begin
 if (FIsRead <> Value) then begin
   FIsRead := Value;
   Database.MarkRead (DocId, Value);
 end;
end;

(******************************************************************************)
procedure TNotesDocument.SetItemByNum(ItemNum: integer; Value: string);
var
 ansiValue: AnsiString;
begin
 ansiValue := Native2Lmbcs(AnsiString(Value));
 CheckError (MailAddHeaderItem(Handle, ItemNum, PAnsiChar(ansiValue), WORD(length(ansiValue))));
end;

//***************************************************
procedure TNotesDocument.ReloadFields;
var
 TempList: TStringList;
begin
 TempList := TStringList.create;
 try
   CheckError( NSFItemScan(Handle, NSFItemScanProc, Pointer(TempList)) );
   FFields.Assign (TempList);
 finally
   TempList.Free;
 end;
end;

(******************************************************************************)
function TNotesDocument.GetRecipients: String;
var
 error: integer;
 fldLength: word;
 ansiRet: AnsiString;
begin
 ansiRet := '';
 SetLength(ansiRet,strLength);
 error := MailGetMessageItem(Handle, MAIL_RECIPIENTS_ITEM_NUM, PAnsiChar(ansiRet), strLength,@fldLength);
 Result := String(System.AnsiStrings.StrPas(PAnsiChar(ansiRet)));
 if error<>0 then
   Result := '';
end;

(******************************************************************************)
procedure TNotesDocument.SetRecipients(Value: string);
var
 hRecipientsList: integer;
 plistRecipients: ptrLIST;
 wRecipientsSize: word;
 n: integer;
 ansiStr, ansiVal: AnsiString;
begin
 if Value = '' then exit;
 CheckError (ListAllocate(0,0,True,@hRecipientsList,@plistRecipients,@wRecipientsSize));
 try
   OSUnlockObject(hRecipientsList);
   n := Pos (',', Value);
   while n > 0 do begin
     ansiStr := Native2Lmbcs(AnsiString(Trim(copy(Value, 1, n-1))));
     Delete (Value, 1, n);
     n := Pos (',', Value);
     CheckError (ListAddEntry(hRecipientsList, True, @wRecipientsSize, 0,
        PAnsiChar(ansiStr),WORD(length(ansiStr))));
   end;
   ansiVal := Native2Lmbcs(AnsiString(Value));
   if ansiVal <> '' then
     CheckError (ListAddEntry(hRecipientsList, True, @wRecipientsSize, 0,
                   PAnsiChar(ansiVal),WORD(length(ansiVal))));
   CheckError (MailAddRecipientsItem (Handle, hRecipientsList, wRecipientsSize));
 except
   OsMemFree (hRecipientsList);
   raise;
 end;
end;

(******************************************************************************)
function TNotesDocument.ReplaceItemValue(itemName: string; value: Variant): TNotesItem;
begin
 Result := nil;
 try
   DeleteItem(ItemName);
 except
 end;
 Items[ItemName].Value := Value;
end;

(******************************************************************************)
procedure TNotesDocument.Save;
var
 uname: string;
 ansiPName, ansiUName: AnsiString;
begin
 if ProfileName <> '' then begin
   // Profile document
   uname := Database.LmbcsUserName;
   if not TNotesName.IsCanonical(uName) then
     uName := TNotesName.TranslateName (uName, True, '');
   ansiUName := AnsiString(uname);
   ansiPName := Native2LMBCS(AnsiString(ProfileName));
   CheckError(NSFProfileUpdate(Handle,PAnsiChar(ansiPName),length(ansiPName),
                  PAnsiChar(ansiUName),length(ansiUName)));
 end
 else begin
   // Normal document
   CheckError(NSFNoteUpdate(Handle,UPDATE_NOCOMMIT));
   if DocId = 0 then
     NSFNoteGetInfo(Handle, _NOTE_ID, @FId);
 end;
end;

(******************************************************************************)

procedure TNotesDocument.Send(fAttachForm: boolean = False; ARecipients: string = '');
var
 hMailBox, hOrigDB: integer;
 szMailBoxPath: array [0..256] of AnsiChar;
 OrigNoteID:longint;
 OrigNoteOID, NewNoteOID: OID;
 tdDate: TIMEDATE;
 sSendTo, sCopyTo, sBCopyTo, sRecip: string;
begin
 if fAttachForm then begin
   if Form = '' then attachForm(MAIL_MEMO_FORM) else attachForm(Form);
 end
 else if Form = '' then
   Form := MAIL_MEMO_FORM;
 if MailFrom = '' then
   MailFrom := Database.UserName;
 if aRecipients <> '' then
   Recipients := aRecipients
 else if Recipients = '' then begin
   // Construct recipients list
   sSendTo := SendTo;
   sCopyTo := CopyTo;
   sBCopyTo := BlindCopyTo;
   if sSendTo = '' then
     sSendTo := Items['SendTo'].AsString;
   if sCopyTo = '' then
     sCopyTo := Items['CopyTo'].AsString;
   if sBCopyTo = '' then
     sBCopyTo := Items['BlindCopyTo'].AsString;
   sRecip := sSendTo;
   if sCopyTo <> '' then
     sRecip:=sRecip+',' + sCopyTo;
   if sBCopyTo <> '' then
     sRecip:=sRecip+',' + sBCopyTo;
   Recipients := sRecip;
 end;

 OSCurrentTIMEDATE(@tdDate);
 CheckError (MailAddHeaderItem(Handle, MAIL_POSTEDDATE_ITEM_NUM,
                PAnsiChar(@tdDate), WORD(sizeof(TIMEDATE))));
 Items['Principal'].asString := MailFrom;

  // Open mailbox
  hMailBox := 0;
  CheckError (OSPathNetConstruct('', PAnsiChar(AnsiString(Database.MailServer)),
                      AnsiString(MAILBOX_NAME), szMailBoxPath));
  CheckError (NSFDbOpen(szMailBoxPath, @hMailBox));
  if (hMailBox=0) then
    raise ELotusNotes.CreateErr(0,'User cancelled Notes session');
  try
    // Take note params
    NSFNoteGetInfo(Handle, _NOTE_DB,  @hOrigDB);
    NSFNoteGetInfo(Handle, _NOTE_ID,  @OrigNoteID);
    NSFNoteGetInfo(Handle, _NOTE_OID, @OrigNoteOID);

    // Set the message's OID database ID to match the mail box */}
    CheckError (NSFDbGenerateOID (hMailBox, @NewNoteOID));
    NSFNoteSetInfo(Handle, _NOTE_DB,  @hMailBox);
    NSFNoteSetInfo(Handle, _NOTE_ID, nil{0});
    NSFNoteSetInfo(Handle, _NOTE_OID, @NewNoteOID);

    // Update message into MAIL.BOX on mail server.}
    CheckError (NSFNoteUpdate(Handle, UPDATE_NOCOMMIT));

    // restore msg to user's mail file and Update to save it there.*/}
    if Database.SaveMail then begin
      NSFNoteSetInfo(Handle, _NOTE_DB, @hOrigDB);
      NSFNoteSetInfo(Handle, _NOTE_ID, @OrigNoteID);
      NSFNoteSetInfo(Handle, _NOTE_OID, @OrigNoteOID);
      CheckError(NSFNoteUpdate(Handle, UPDATE_NOCOMMIT));
    end;
  finally
    NSFDbClose (hMailBox);
  end;
end;

(******************************************************************************)
function TNotesDocument.Attach(AName: string): integer;
begin
 if not fileExists(AName) then
   raise ELotusNotes.Create('File to be attached does not exist: ' + aName);
 // Attach the file
 CheckError(MailAddMessageAttachment (Handle, PAnsiChar(Native2Lmbcs(AnsiString(AName))),
                 PAnsiChar(Native2Lmbcs(AnsiString(ExtractFileName(AName))))));
 Result := 0;
end;

(******************************************************************************)
procedure TNotesDocument.Detach(Index: integer; FileName: string);
var
 SourceName: string;
 atcItemcl: BLOCKIDcl;
begin
 SourceName := FAttach[Index];
 atcItemcl := BLOCKIDcl(FAttach.Objects[Index]);
 if FileName = '' then
   FileName := SourceName;
 CheckError(MailExtractMessageAttachment(Handle, atcItemcl.BlockItem,
               PAnsiChar(Native2Lmbcs(AnsiString(FileName)))));
end;

//**********************************************
function TNotesDocument.GetFieldCount: Integer;
begin
 Result := FFields.count;
end;

//**********************************************
function TNotesDocument.GetFieldName(Index: integer): string;
begin
 Result := FFields[Index];
end;

//**********************************************
function TNotesDocument.GetItemByName(ItemName: string): TNotesItem;
var
 Index: integer;
 ansiName2: AnsiString;
begin
 Index := FFields.indexOf(ItemName);
 if Index < 0 then begin
   Result := TNotesItem.createNew (self, ItemName);
   FFields.addObject (ItemName, Result);
 end
 else begin
   if FFields.Objects[Index] = nil then begin
     ansiName2 := Native2Lmbcs(AnsiString(ItemName));
     if NSFItemIsPresent (Handle, PAnsiChar(ansiName2), Length(ansiName2)) then
       FFields.Objects[Index] := TNotesItem.create(self, ItemName)
     else begin
       FFields.Objects[Index] := TNotesItem.createNew(self, ItemName);
     end;
   end;
   Result := FFields.Objects[Index] as TNotesItem;
 end;
end;

//****************************************************
constructor TNotesDocument.CreateProfile(notesDatabase: TNotesDatabase; aProfileName: string);
var
 anID: NOTEID;
 uname: string;
 ansiUNname, ansiPName: AnsiString;
begin
 inherited Create;
 uname := notesDatabase.LmbcsUserName;
 if not TNotesName.IsCanonical(uName) then
   uName := TNotesName.TranslateName (uName, True, '');
 ansiUNname := AnsiString(uname);
 ansiPName := Native2LMBCS(AnsiString(aProfileName));
 CheckError(NSFProfileOpen(notesDatabase.Handle,PAnsiChar(ansiPName),Length(ansiPName),
                 PAnsiChar(ansiUNname),Length(ansiUNname), TRUE, anID));
 FProfileName := aProfileName;
 InitDocument (NotesDatabase, anID);
end;

//****************************************************
constructor TNotesDocument.CreateResponse(notesDatabase: TNotesDatabase;  MainDoc: TNotesDocument);
begin
 CreateNew(notesDatabase);
 Items['$REF'].AsReference := MainDoc.UniversalID;
 if MainDoc.IsItemExists('Subject') then
   Subject := 'Re: ' + MainDoc.Subject;
end;

//**********************************************
constructor TNotesDocument.CreateResponseByUNID(notesDatabase: TNotesDatabase; const anUNID: UNID);
begin
 CreateNew(notesDatabase);
 Items['$REF'].AsReference := anUNID;
end;

//**********************************************
procedure TNotesDocument.DeleteItem(ItemName: string);
var
 Index: integer;
 ansiName2: AnsiString;
begin
 ansiName2 := Native2Lmbcs(AnsiString(ItemName));
 while NsfItemDelete(Handle, PAnsiChar(ansiName2), length(ansiName2)) = 0 do;
 Index := FFields.indexOf(ItemName);
 if Index >= 0 then begin
   FFields.Objects[Index].free;
   FFields.delete (Index);
 end;
end;

//****************************************************
function TNotesDocument.GetOriginatorID: OID;
var
 td: TIMEDATE;
 cls: word;
begin
 CheckError( NSFDbGetNoteInfo(Database.Handle, DocID, @Result, @td, @cls) );
end;

//****************************************************
function TNotesDocument.GetUniversalID: UNID;
var
 id: OID;
begin
 id := GetOriginatorID;
 Result.aFile := id.FileNum;
 Result.Note := id.Note;
end;

//***************************************************
procedure TNotesDocument.InitDocument(notesDatabase: TNotesDatabase;  anID: DWORDapi);
var
 atcNum: integer;
 atcItem: BLOCKID;
 atcItemcl: BLOCKIDcl;
 ansiAtcFileName: AnsiString;
begin
  FFields := TStringList.create;
  ReloadFields;
  FDatabase := notesDatabase;
  FAttach := TStringList.Create;
  fId := anId;
  FSummary := TStringList.create;

  // Load attachments list
  atcNum := 0;
  while True do begin
    setLength (ansiAtcFileName, 256);
    if not MailGetMessageAttachmentInfo (Handle, atcNum, @atcItem, PAnsiChar(ansiAtcFileName),
                     nil,nil,nil,nil,nil) then break;
    ansiAtcFileName := System.AnsiStrings.strPas(PAnsiChar(ansiAtcFileName));
    if ansiAtcFileName <> '' then begin
      atcItemcl := BLOCKIDcl.Create;
      atcItemcl.BlockItem := atcItem;
      FAttach.AddObject(String(AnsiUpperCase(Lmbcs2Native(ansiAtcFileName))), atcItemcl);
      inc(atcNum);
    end;
  end;
end;

//****************************************************
function TNotesDocument.IsItemExists(ItemName: string): boolean;
var
 ansiName2: AnsiString;
begin
 ansiName2 := Native2Lmbcs(AnsiString(ItemName));
 Result := NSFItemIsPresent (Handle, PAnsiChar(ansiName2), length(ansiName2));
end;

//******************************************************************************
function TNotesDocument.GetAttachmentCount: Integer;
begin
 Result := FAttach.count;
end;

//******************************************************************************

function TNotesDocument.GetAttachment(Index: integer): string;
begin
  Result := FAttach[Index];
end;

//******************************************************************************

procedure TNotesDocument.SaveAttachment(Index: integer; FilePath: String);
var
 blid: BLOCKIDcl;
 ansiFilePath: AnsiString;
begin
 if (Index < 0) or (Index >= FAttach.count) then
   exit;
 ansiFilePath   := AnsiString(FilePath);
 blid := BLOCKIDcl(FAttach.Objects[Index]);
 CheckError( NSFNoteExtractFile(Handle, blid.blockItem, PAnsiChar(ansiFilePath), Nil));
end;

//******************************************************************************

function TNotesDocument.GetLastModified: TDateTime;
var
  T: TIMEDATE;
begin
  NsfNoteGetInfo(Handle, _NOTE_MODIFIED, @T);
  Result := NotesToDateTime(T);
end;

//******************************************************************************
function TNotesDocument.GetLastAccessed: TDateTime;
var
  T: TIMEDATE;
begin
  NsfNoteGetInfo (Handle, _NOTE_ACCESSED, @T);
  Result := NotesToDateTime (T);
end;

//******************************************************************************

function TNotesDocument.Evaluate(aFormula: string): variant;
var
 hFCompiled: FormulaHandle;
 hFComputed: HCOMPUTE;
 wFormulaErr: STATUS;
 wFormulaLen, wFormulaErrLine,
 wFormulaErrCol, wFormulaErrOffset, wFormulaErrLen: word;
 hResult: LHANDLE;
 wResultLen: word;
 fNoteMatchesFormula, fNoteToDelete, fNoteModified: boolean;
 phFCompiled, phResult: pointer;
 ansiFormula: AnsiString;
begin
 // Formula compilation
 hFCompiled := 0;
 ansiFormula := AnsiString(aFormula);
 try
   CheckError (NSFFormulaCompile(nil,0,PAnsiChar(ansiFormula),Length(ansiFormula),@hFCompiled,
        @wFormulaLen, @wFormulaErr, @wFormulaErrLine, @wFormulaErrCol, @wFormulaErrOffset,
        @wFormulaErrLen));
 except
   on E: ELotusNotes do begin
     raise ELnFormulaCompile.CreateErr(E.ErrorCode, E.Message,
                 wFormulaErr, wFormulaErrOffset, wFormulaErrLen);
   end
   else
     raise;
 end;

 // Evaluation
 hFComputed := 0;
 hResult := 0;
 try
   phFCompiled := OSLockObject (hFCompiled);
   CheckError (NSFComputeStart (0, phFCompiled, @hFComputed));
   CheckError (NSFComputeEvaluate (hFComputed, self.Handle, @hResult, @wResultLen,
      @fNoteMatchesFormula, @fNoteToDelete, @fNoteModified));
   phResult := OsLockObject (hResult);
   Result := BufferToValue(phResult, wResultLen);
 finally
   if hResult <> 0 then begin
     OsUnlockObject (hResult);
     OsMemFree (hResult);
   end;
   if hFCompiled <> 0 then begin
     OsUnlockObject (hFCompiled);
     OsMemFree (hFCompiled);
   end;
   if hFComputed <> 0 then
     NSFComputeStop(hFComputed);
 end;
end;

//******************************************************************************

function TNotesDocument.FindAttachment(aName: string): integer;
begin
 Result := FAttach.indexOf(AnsiUpperCase(aName));
end;

//******************************************************************************

procedure TNotesDocument.DeleteAttachment(Index: integer);
var
 blid: BLOCKIDcl;
begin
 if (Index < 0) or (Index >= FAttach.count) then
   exit;
 blid := BLOCKIDcl(FAttach.Objects[Index]);
 CheckError(NSFNoteDetachFile(Handle, blid.blockItem));
 FAttach.Objects[Index].free;
 FAttach.delete(Index);
end;

//******************************************************************************

procedure TNotesDocument.LoadFontTable;
type
 PCDFONTTABLE = ^CDFONTTABLE;
 PCDFACE = ^CDFACE;
var
 bhValue: BlockID;
 szValue: DWORDapi;
 pTbl: PCDFONTTABLE;
 pFace: PCDFACE;
 n: integer;
 FontID: DWORDapi;
 ff: word;
begin
 if FFontTable <> nil then
   exit; //alredy loaded

  // Font table resides in $FONTS item of TYPE_COMPOSITE
 FFontTable := TStringList.create;
 FMaxFontID := STATIC_FONT_FACES-1;
 if NsfItemInfo (Handle, ITEM_NAME_FONTS, length(ITEM_NAME_FONTS), nil, nil,
                 @bhValue, @szValue) <> NOERROR then
   exit; //no such item
 pTbl := PCDFONTTABLE(DWORDapi(OsLockBlock(bhValue)));
 pTbl := PCDFONTTABLE(DWORDapi(pTbl) + sizeOf(word));
 try
   pFace := PCDFACE(DWORDapi(pTbl) + sizeOf(CDFONTTABLE));
   for n := 1 to pTbl^.Fonts do begin
     FontID := MakeLong (pFace^.Face, pFace^.Family);
     FFontTable.AddObject(String(Lmbcs2Native(System.AnsiStrings.strPas(pFace^.Name))),
                          pointer(FontID));
     ff := pFace^.Face and $FF;
     if ff > FMaxFontId then
       FMaxFontId := ff;
     pFace := PCDFACE (DWORDapi(pFace) + ODSLength(_CDFACE));
   end;
 finally
   OsUnlockBlock(bhValue);
 end;
end;

//***************************************************
procedure TNotesDocument.SaveFontTable;
type
 PCDFONTTABLE = ^CDFONTTABLE;
 PCDFACE = ^CDFACE;
var
 MemSz: DWORDapi;
 pBuf: pointer;
 pTbl: PCDFONTTABLE;
 pFace: PCDFACE;
 FontId: DWORDapi;
 i: integer;
 ansiNameFonts: AnsiString;
begin
 if (FFontTable = nil) or (FFontTable.count = 0) then
   exit;
 // Generate CD buffer
 MemSz := sizeOf(CDFONTTABLE) + sizeOf(CDFACE) * FFontTable.count;
 if (MemSz mod 2) <> 0 then
   inc(MemSz);
 GetMem(pBuf, MemSz);
 try
   // Fill header
   pTbl := PCDFONTTABLE(pBuf);
   pTbl^.Header.Signature := SIG_CD_FONTTABLE;
   pTbl^.Header.Length := MemSz;
   pTbl^.Fonts := FFontTable.count;

   // Add faces
   pFace := PCDFACE(DWORDapi(pBuf) + sizeOf(CDFONTTABLE));
   for i := 0 to FFontTable.count-1 do begin
     System.AnsiStrings.strCopy (pFace^.Name, PAnsiChar(Native2Lmbcs(AnsiString(FFontTable[i]))));
     FontId := DWORDapi(FFontTable.Objects[i]);
     pFace^.Face := LoWord (FontId);
     pFace^.Family := HiWord (FontId);
     pFace := PCDFACE(DWORDapi(pFace) + sizeOf(CDFACE));
   end;

   // Add item
   ansiNameFonts := AnsiString(ITEM_NAME_FONTS);
   NsfItemDelete (Handle, PAnsiChar(ansiNameFonts), length(ansiNameFonts));
   CheckError (NSFItemAppend (Handle, 0, PAnsiChar(ansiNameFonts), length(ansiNameFonts),
                              TYPE_COMPOSITE, pBuf, MemSz));
 finally
   FreeMem(pBuf);
 end;
end;

//**********************************************
procedure TNotesDocument.Sign;
begin
 CheckError (NsfNoteSign (Handle));
end;

//**********************************************
function TNotesDocument.GetSignature(var SignedBy:    string;
                                     var CertifiedBy: string;
                                     pTime: PTDateTime): boolean;
var
 d: TIMEDATE;
 ansiSignedBy, ansiCertifiedBy: AnsiString;
begin
 SetLength(ansiSignedBy, MAXUSERNAME+1);
 SetLength(ansiCertifiedBy, MAXUSERNAME+1);
 Result := NsfNoteVerifySignature (Handle, nil, @d,
                   PAnsiChar(ansiSignedBy), PAnsiChar(ansiCertifiedBy)) = NOERROR;
 if Result then begin
   SignedBy    := String(System.AnsiStrings.strPas(PAnsiChar(ansiSignedBy)));
   CertifiedBy := String(System.AnsiStrings.strPas(PAnsiChar(ansiCertifiedBy)));
   if pTime <> nil then
     pTime^ := NotesToDateTime(d);
 end;
end;

//***************************************************
procedure TNotesDocument.UpdateUnread;
var
 uName: string;
 ansiUName: AnsiString;
 hTable: THandle;
begin
 // Get IDs
 uName := Database.LmbcsUserName;
 if not TNotesName.IsCanonical(uName) then
   uName := TNotesName.TranslateName(uName, True, '');
 ansiUName := AnsiString(uName);
 hTable := 0;
 CheckError (NSFDbGetUnreadNoteTable(Database.Handle, PAnsiChar(ansiUName), length(ansiUName), True, @hTable));
 try
   // Check ID
   fIsRead := not (IDIsPresent(hTable, DocID));
 finally
   if hTable <> 0 then
     OsMemFree (hTable);
 end;
end;

//******************************************************************************

function TNotesDocument.CountMultipleItems(aName: string): integer;
var
 item_new, item_cur: TNotesItem;
 first_item: boolean;
begin
 Result := 0;
 first_item := True;
 item_cur := Items[aName];
 if not item_cur.IsNewItem then
   exit;
 inc(Result);
 while item_cur.NextItemExists do begin
   try
     item_new := item_cur.CreateNextItem;
   finally
     if not first_item then item_cur.free;
   end;
   inc(Result);
   item_cur := item_new;
   first_item := False;
 end;
end;

//***************************************************
function TNotesDocument.LoadMultipleItems(aName: string): TList;
var
  item_new, item_cur: TNotesItem;
begin
  Result := TList.create;
  try
    item_cur := Items[aName];
    if not item_cur.IsNewItem then exit;
    Result.add(item_cur);

    while item_cur.NextItemExists do begin
      item_new := item_cur.CreateNextItem;
      Result.add(item_new);
      item_cur := item_new;
    end;
  except
    Result.free;
    raise;
  end;
end;

//ClassMarker_Method(TNotesDocument)


//***************************************************
// TNotesDirectory
//***************************************************
constructor TNotesDirectory.Create;
begin
  inherited Create;
  FPorts := TStringList.create;
end;

//***************************************************
destructor TNotesDirectory.Destroy;
begin
  FindClose;
  FPorts.free;
  inherited Destroy;
end;

//***************************************************
procedure TNotesDirectory.FindClose;
var
  i: integer;
begin
  if SrcTable <> nil then begin
    for i := 0 to SrcTable.count-1 do FreeMem(SrcTable[i]);
    SrcTable.free;
    SrcTable := nil;
  end;
  if hDirectory <> 0 then begin
    NSFDbClose(hDirectory);
    hDirectory := 0;
  end;
end;

(******************************************************************************)
function OnDirFound (Obj: pointer;
                     search_info: pSEARCH_MATCH;
                     summary_info: pITEM_TABLE): STATUS; far; stdcall;
var
 pItems: pITEM;
 pValues: PAnsiChar;
 i, n: integer;
 pEntry: PNotesDirEntry;
 ansiName, ansiValue: AnsiString;
 {pValues, valptr, pBuf: PAnsiChar;
 i, j, n: integer;
 pEntry: PNotesDirEntry;
 sName, sValue, buf: string;
 wType, bufSz: word;
 lst: pList;}
 val: variant;
begin
 if ((search_info.SERetFlags and SE_FMATCH) <> 0) then begin
   // summary_info contains a list of items describing directory entry found
   // First goes ITEM_TABLE, Items field contains total number of entries
   // Next goes array of ITEM
   pItems := pItem (DWORDapi (summary_info) + sizeOf(ITEM_TABLE));

   // Next goes array of records <name><type><value>
   // name is Item.NameLength, value is Item.ValueLength, type is WORD
   pValues := PAnsiChar (DWORDapi (summary_info) + sizeOf(ITEM_TABLE) + summary_info^.Items * sizeOf(ITEM));

   GetMem (pEntry, sizeOf(TNotesDirEntry));
   try
     FillChar (pEntry^, sizeOf(TNotesDirEntry), 0);

     // For each item...
     for i := 1 to summary_info^.Items do begin
       ansiValue:= '';
       ansiName := '';
       setLength (ansiName,  pItems^.NameLength+2);
       setLength (ansiValue, pItems^.ValueLength+2);
       // Get name
       System.AnsiStrings.strLCopy (PAnsiChar(ansiName), pValues, pItems^.NameLength);
       ansiName[pItems^.NameLength+1] := #0;
       ansiName := Lmbcs2Native (System.AnsiStrings.strPas(PAnsiChar(ansiName)));
       // Get value
       val := BufferToValue(PAnsiChar(DWORDapi(pValues) + pItems^.NameLength), pItems^.ValueLength);
       try
         ansiValue := AnsiString(VarAsType(val, varString));
       except
         ansiValue := '';
       end;
       if System.AnsiStrings.compareText(ansiName,AnsiString('$title')) = 0 then begin
         //-- pEntry^.FileName := System.AnsiStrings.trim(ansiValue);
         System.AnsiStrings.strLCopy(pEntry^.FileName,
                                     PAnsiChar(System.AnsiStrings.trim(ansiValue)),
                                     MAXPATH-1);
       end
       else if System.AnsiStrings.compareText(ansiName,AnsiString('$type')) = 0 then begin
         pEntry^.EntryType := System.AnsiStrings.compareText(
                                     System.AnsiStrings.trim(ansiValue),AnsiString('$dir')) = 0;
       end
       else if System.AnsiStrings.compareText(ansiName,AnsiString('$info')) = 0 then begin
         n := Pos (AnsiString('#'), ansiValue); //template name follows #
         if n = 0 then begin
           //-- pEntry^.FileInfo := System.AnsiStrings.Trim(ansiValue);
           System.AnsiStrings.strLCopy(pEntry^.FileInfo,
                                       PAnsiChar(System.AnsiStrings.trim(ansiValue)),
                                       MAXPATH-1);
         end
         else begin
           //-- pEntry^.FileInfo := System.AnsiStrings.Trim(Copy(ansiValue, 1, n-2));
           System.AnsiStrings.strLCopy(pEntry^.FileInfo,
                                       PAnsiChar(System.AnsiStrings.trim(Copy(ansiValue, 1, n-2))),
                                       MAXPATH-1);
         end;
       end;
       pValues := PAnsiChar(DWORDapi(pValues) + pItems^.NameLength + pItems^.ValueLength);
       pItems  := pItem(DWORDapi(pItems) + sizeOf(ITEM_TABLE));
     end;
     if pEntry^.FileName = '' then
       FreeMem(pEntry)
     else
       TList(Obj).Add(pEntry);
   except
     FreeMem (pEntry);
     raise;
   end;
 end;
 Result := NOERROR;
end;

//***************************************************
function CompareProc (Item1, Item2: Pointer): Integer;
begin
 Result := -1 * (ord(pNotesDirEntry(Item1)^.EntryType) - ord(pNotesDirEntry(Item2)^.EntryType));
 if Result = 0 then begin
   if pNotesDirEntry(Item1)^.EntryType then
     Result := System.AnsiStrings.compareText (pNotesDirEntry(Item1)^.FileName, pNotesDirEntry(Item2)^.FileName)
   else
     Result := System.AnsiStrings.compareText (pNotesDirEntry(Item1)^.FileInfo, pNotesDirEntry(Item2)^.FileInfo);
 end;
end;

//***************************************************
function TNotesDirectory.FindFirst(const Server, Path: string;
                             Options: TNotesFindOptions;
                             var Entry: TNotesDirEntry): boolean;
var
 ansiPath, ansiSrcServer, ansiSrcPath: AnsiString;
 error: STATUS;
 wFileType: word;
begin
 FindClose;
 if (CompareText (Server, 'Local') = 0) then
   SrcServer := ''
 else
   SrcServer := Server;
 SrcPath := Path;
 SrcOptions := Options;
 if SrcOptions = [] then
   SrcOptions := [nfoFiles];

 // Open a directory handle
 ansiSrcServer := AnsiString(SrcServer);
 ansiSrcPath   := AnsiString(SrcPath);;
 ansiPath := '';
 SetLength (ansiPath, 256);
 CheckError ( OSPathNetConstruct('', PAnsiChar(ansiSrcServer),
                  PAnsiChar(ansiSrcPath), PAnsiChar(ansiPath)) );
 ansiPath := Native2Lmbcs (System.AnsiStrings.StrPas(PAnsiChar(ansiPath)));

 error := NSFDbOpen(PAnsiChar(ansiPath), @hDirectory);
 if ((hDirectory=0) and (Error=0)) or (Error=USER_CANCEL) then
   raise ELotusNotes.CreateErr(0,'User cancelled Notes session');
 CheckError (Error);

 // Enumeration
 wFileType := 0;
 if (nfoFiles in SrcOptions) then
   wFileType := FILE_DBANY
 else if (nfoTemplates in SrcOptions) then
   wFileType := FILE_FTANY;
 if (nfoSubDirs in SrcOptions) then
   wFileType := wFileType or FILE_DIRS;
 wFileType := wFileType or FILE_NOUPDIRS;
 SrcTable := TList.create;
 try
   // List all entries
   CheckError (NSFSearch (
      hDirectory,         //* directory handle           */
      0,                  //* selection formula          */
      nil,                //* title of view in formula   */
      SEARCH_FILETYPE +   //* search for files           */
      SEARCH_SUMMARY,     //* return a summary buffer    */
      wFileType,          // file type
      nil,                //* starting date              */
      OnDirFound,         //* call for each file found   */
      SrcTable,           //* argument to action routine */
      nil));              //* returned ending date (unused) */

   // Sort the list
   SrcTable.Sort(CompareProc);

   // Return first one
   SrcIndex := 0;
   Result := FindNext(Entry);
 except
   FindClose;
   raise;
 end;
end;

//***************************************************
function TNotesDirectory.FindNext(var Entry: TNotesDirEntry): boolean;
begin
 if SrcIndex >= SrcTable.count then begin
   Result := False;
   FindClose;
 end
 else begin
   Result := True;
   Entry := PNotesDirEntry(SrcTable[SrcIndex])^;
   inc(SrcIndex);
 end;
end;

//***************************************************
function TNotesDirectory.GetPorts: TStrings;
var
 ansiStr: AnsiString;
 n: integer;
begin
 if FPorts.count = 0 then begin
   // Read list of ports from NOTES.INI
   ansiStr := '';
   SetLength(ansiStr,256);
   OSGetEnvironmentString (PAnsiChar(AnsiString('Ports')), PAnsiChar(ansiStr), 255);
   ansiStr := System.AnsiStrings.StrPas(PAnsiChar(ansiStr));
   // Entries are separated by comma
   n := Pos (AnsiString(','), ansiStr);
   while n <> 0 do begin
     FPorts.add(String(System.AnsiStrings.trim(copy(ansiStr, 1, n-1))));
     delete(ansiStr, 1, n);
     n := Pos(AnsiString(','), ansiStr);
   end;
   if ansiStr <> '' then
     FPorts.add(String(System.AnsiStrings.Trim(copy(ansiStr, 1, n-1))));
 end;
 Result := FPorts;
end;

//***************************************************
procedure TNotesDirectory.ListServers(Port: string; List: TStrings);
var
 hList:  HANDLE;
 pcName: PAnsiChar;
 pList:  pointer;
 i, wCount: word;
 pwLengths: pword;
 pcNames: PAnsiChar;
 ansiBuf: AnsiString;
 sName: String;
begin
 if Port = '' then
   pcName := nil
 else
   pcName := PAnsiChar(AnsiString(Port));
 CheckError( NSGetServerList(pcName,@hList) );
 List.clear;
 List.add('Local');
 pList := OSLockObject (hList);
 try
   // the list contains counter (word), list of name lengths and list of names
   wCount    := pword(pList)^;
   pwLengths := pword (DWORDapi(pList)     + sizeOf(wCount));
   pcNames   := PAnsiChar (DWORDapi(pList) + sizeOf(wCount) + wCount*sizeOf(word));
   for i := 1 to wCount do begin
     ansiBuf := '';
     setLength (ansiBuf, pwLengths^ + 2);
     System.AnsiStrings.strLCopy (PAnsiChar(ansiBuf), pcNames, pwLengths^);
     ansiBuf[pwLengths^+1] := #0;
     ansiBuf := Lmbcs2Native (System.AnsiStrings.StrPas(PAnsiChar(ansiBuf)));
     sName := String(ansiBuf);
     sName := TNotesName.TranslateName(sName, False, '');
     List.add(sName);
     pcNames   := PAnsiChar(DWORDapi(pcNames) + pwLengths^);
     pwLengths := pword (DWORDapi(pwLengths)  + sizeOf(word));
   end;
 finally
   OsUnlockObject(hList);
   OsMemFree(hList);
 end;
end;
//ClassMarker_Method(TNotesDirectory)


//******************************************************************************
// TNotesName
//******************************************************************************

const
  MAX_PARTS = 12;
  PartLabels: array [0..MAX_PARTS] of string = (
    'CN', 'G', 'S', 'I', 'Q', 'C', 'OU', 'OU', 'OU', 'OU', 'ADMD', 'PRMD', 'O'
  );

//******************************************************************************

class function TNotesName.CheckAddress(aServer, aName: string): boolean;
var
 Values: TStringList;
begin
 Values := TStringList.create;
 try
   Result := LookupName(aServer, [USER_NAMESSPACE], [aName], ['FullName'], [nloExhaustive], Values);
 finally
   Values.free;
 end;
end;

//***************************************************
constructor TNotesName.Create(aName: String);
begin
 inherited Create;
 SetName (aName);
end;

//***************************************************
function TNotesName.GetKeyword;
begin
 Result := Format('%s/%s/%s/%s/%s/%s', [Country, Organization, OrgUnit1, OrgUnit2, OrgUnit3, OrgUnit4]);
end;

//******************************************************************************

function TNotesName.GetNameComponent(Index: integer): string;

  procedure getstr (var Str: AnsiString; const buf: PAnsiChar; const sz: word);
  begin
    if buf = nil then
      Str := ''
    else begin
      SetLength(Str, sz + 1);
      System.AnsiStrings.strLCopy (PAnsiChar(Str), buf, sz);
      Str[sz+1] := #0;
      Str := System.AnsiStrings.StrPas(PAnsiChar(Str));
    end;
  end;

var
 ansiRet: AnsiString;
begin
 if not fParsed then begin
   CheckError (DnParse (0, nil, PAnsiChar(AnsiString(FName)), FComponents, sizeOf(FComponents)));
   fParsed := True;
 end;
 case Index of
   0: begin
     getstr(ansiRet, FComponents.CN, FComponents.CNLength);
     if ansiRet = '' then
       ansiRet := AnsiString(FName);
   end;
    1:  getstr(ansiRet, FComponents.G,     FComponents.GLength);
    2:  getstr(ansiRet, FComponents.S,     FComponents.SLength);
    3:  getstr(ansiRet, FComponents.I,     FComponents.ILength);
    4:  getstr(ansiRet, FComponents.Q,     FComponents.QLength);
    5:  getstr(ansiRet, FComponents.C,     FComponents.CLength);
    6:  getstr(ansiRet, FComponents.OU[0], FComponents.OULength[0]);
    7:  getstr(ansiRet, FComponents.OU[1], FComponents.OULength[1]);
    8:  getstr(ansiRet, FComponents.OU[2], FComponents.OULength[2]);
    9:  getstr(ansiRet, FComponents.OU[3], FComponents.OULength[3]);
    10: getstr(ansiRet, FComponents.ADMD,  FComponents.ADMDLength);
    11: getstr(ansiRet, FComponents.PRMD,  FComponents.PRMDLength);
    12: getstr(ansiRet, FComponents.O,     FComponents.OLength);
  end;
 Result := String(ansiRet);
end;

//******************************************************************************

class function TNotesName.IsCanonical(aName: string): boolean;
var
 i: integer;
begin
 // Tries to find any of named part labels
 Result := True;
 aName := UpperCase(aName);
 for i := System.Low(PartLabels) to System.High(PartLabels) do begin
   if Pos ('/' + PartLabels[i] + '=', aName) > 0 then
     exit;
 end;
 Result := False;
end;

//******************************************************************************

function TNotesName.IsHirerarchical: boolean;
var
 n: integer;
begin
 n := Pos ('/', FName);
 Result := (n > 0) and (n < length(FName));
end;

//******************************************************************************

class function TNotesName.LookupName(ServerName: string;
                     NameSpaces: array of string;
                     Names: array of string;
                     Items: array of string;
                     Flags: TNotesLookupOptions;
                     Values: TStrings): boolean;
var
 lNameSpaces, lNames, lItems: TStringList;
 i: integer;
begin
 lNameSpaces := TStringList.create;
 lNames := TStringList.create;
 lItems := TStringList.create;
 try
   for i := System.Low(NameSpaces) to System.High(NameSpaces) do
     lNameSpaces.add(NameSpaces[i]);
   for i := System.Low(Names) to System.High(Names) do
     lNames.add(Names[i]);
   for i := System.Low(Items) to System.High(Items) do
     lItems.add(Items[i]);
   Result := LookupNameList(ServerName, lNameSpaces, lNames, lItems, Flags, Values);
 finally
   lNameSpaces.free;
   lNames.free;
   lItems.free;
 end;
end;

//***************************************************
class function TNotesName.LookupNameList(ServerName: string;
                             NameSpaces: TStrings;
                             Names: TStrings;
                             Items: TStrings;
                             Flags: TNotesLookupOptions;
                             Values: TStrings): boolean;
const
 LookupMap: array [TNotesLookupOption] of word = (NAME_LOOKUP_ALL, NAME_LOOKUP_NOSEARCHING, NAME_LOOKUP_EXHAUSTIVE);
var
 pcServer: PAnsiChar;
 sNameSpaces, sNames, sItems: PAnsiChar;
 nNameSpaces, nNames, nItems: word;
 wFlags:  word;
 iLookup: TNotesLookupOption;
 hBuffer: LHandle;
 pLookup, pName, pMatch, pItem: PAnsiChar;
 wDataType, wSize, nMatches: word;
 i, j, n, k, items_added: integer;
 ansiBuf: AnsiString;

  procedure GetBuf (Arr: TStrings; var Buf: PAnsiChar; var nBuf: word);
  var
    sz, n, i: integer;
  begin
    sz := 0;
    Buf := nil;
    nBuf := 0;
    for i := 0 to Arr.count-1 do if Arr[i] <> '' then begin
      n := length(Arr[i]);
      ReallocMem (Buf, sz + n + 1);
      System.AnsiStrings.strCopy(PAnsiChar(DWORDapi(Buf) + sz), PAnsiChar(AnsiString(Arr[i])));
      //Buf[sz+n+1] := #0;
      inc (nBuf);
      inc (sz, n + 1);
    end;
  end;

begin
  sNameSpaces := nil;
  sNames := nil;
  sItems := nil;
  try
    // Prepare params
    if (NameSpaces.count = 0) or (NameSpaces[0] = '') then NameSpaces[0] := USER_NAMESSPACE;
    GetBuf (NameSpaces, sNameSpaces, nNameSpaces);
    GetBuf (Names, sNames, nNames);
    GetBuf (Items, sItems, nItems);
    if ServerName = '' then
      pcServer := nil
    else
      pcServer := PAnsiChar(AnsiString(ServerName));
    wFlags := 0;
    for iLookup := System.Low(TNotesLookupOption) to System.High(TNotesLookupOption) do
      if iLookup in Flags then wFlags := wFlags or LookupMap[iLookup];
    if nNames = 0 then begin  //changed by Noah Silva
      nNames := 1;
      sNames := '\0';
    end;
    // Get data
    hBuffer := 0;
    CheckError (NAMELookup(pcServer,
                           wFlags, nNameSpaces,
                           sNameSpaces,
                           nNames, sNames, nItems,
                           sItems, hBuffer));
    // List names
    Values.clear;
    pLookup := OsLockObject (hBuffer);
    pName := nil;
    try
      for i := 1 to nNameSpaces * nNames do begin
        // Get name ptr
        pName := NAMELocateNextName (pLookup, pName, @nMatches);
        if pName = nil then
          break;
        pMatch := nil;
        for j := 0 to nMatches-1 do begin
          // Get match ptr
          items_added := 0;
          pMatch := NAMELocateNextMatch (pLookup, pName, pMatch);
          if pMatch = nil then break;
          for n := 0 to nItems-1 do begin
            // Get item ptr
            pItem := NAMELocateItem (pMatch, n, wDataType, @wSize);
            if pItem = nil then
              continue;   // if item not present in this match, go on to next match
            // Gets value
            k := 0;
            ansiBuf := '';
            setLength (ansiBuf, 255);
            while NAMEGetTextItem (pMatch, n, k, PAnsiChar(ansiBuf), 254) = NOERROR do begin
              Values.add(String(System.AnsiStrings.StrPas(PAnsiChar(ansiBuf))));
              inc(k);
              inc(items_added);
            end;
          end;
          while items_added < nItems do begin
            Values.add('');
            inc(items_added);
          end;
          //Values.add(LOOKUP_VALUE_END);
        end;
      end;
    finally
      OsUnlockObject (hBuffer);
      OsMemFree (hBuffer);
    end;
  finally
    if sNameSpaces <> nil then
      FreeMem(sNameSpaces);
    if sNames <> nil then
      FreeMem(sNames);
    if sItems <> nil then
      FreeMem(sItems);
  end;
  Result := Values.count > 0;
end;

//***************************************************
procedure TNotesName.SetName (Value: string);
begin
 if Value <> fName then begin
   fParsed := False;
   if IsCanonical(Value) then
     fName := Value
   else
     fName := TranslateName(Value, True, '');
 end;
end;

//***************************************************
procedure TNotesName.SetNameComponent(Index: integer; Value: string);
var
 parts: array [0..MAX_PARTS] of string;
 i: integer;
 Canon: string;
begin
 for i := System.Low(parts) to System.High(parts) do
   parts[i] := GetNameComponent(i);
 Canon := '';
 parts[Index] := Value;
 for i := System.Low(parts) to System.High(parts) do begin
   if parts[i] <> '' then begin
     Canon := Canon + '/' + PartLabels[i] + '=' + parts[i];
   end;
 end;
 Canonical := Canon;
end;

//***************************************************
function TNotesName.GetAbbreviatedName: string;
begin
 Result := TranslateName(Canonical, False, TemplateName);
end;

//******************************************************************************

procedure TNotesName.SetAbbreviatedName(Value: string);
begin
 SetName(Value);
end;

//******************************************************************************

class function TNotesName.TranslateName(aName: string;
                  fToCanonical: boolean;
                  aTemplate: string): string;
var
 pct: PAnsiChar;
 sz: word;
 ansiRet: AnsiString;
begin
 sz := MAXUSERNAME;
 SetLength (ansiRet, MAXUSERNAME + 1);
 if aTemplate = '' then
   pct := nil
 else
   pct := PAnsiChar(AnsiString(aTemplate));
 if fToCanonical then
   CheckError (DNCanonicalize(0, pct, PAnsiChar(AnsiString(aName)), PAnsiChar(ansiRet), 255, sz))
 else
   CheckError (DNAbbreviate(  0, nil, PAnsiChar(AnsiString(aName)), PAnsiChar(ansiRet), 255, sz));
 ansiRet := System.AnsiStrings.StrPas(PAnsiChar(ansiRet));
 Result  := String(ansiRet);
end;
//ClassMarker_Method(TNotesName)

//******************************************************************************
// TNotesView
//******************************************************************************
constructor TNotesView.Create(notesDatabase: TNotesDatabase; aName: string);
begin
  inherited Create(notesDatabase);
  FName := aName;
  Load('', 0);
end;

constructor TNotesView.CreateExt(notesDatabase: TNotesDatabase; aName,
  aKey: string; nMaxDocs: integer);
begin
  inherited Create (notesDatabase);
  FName := aName;
  Load(aKey, nMaxDocs);
end;

//****************************************************
constructor TNotesView.CreateEmpty(notesDatabase: TNotesDatabase; aName: string);
begin
 inherited Create(notesDatabase);
 FName := aName;
end;

//****************************************************
function TNotesView.GetIsFolder: Boolean;
begin
 Result := Pos(DESIGN_FLAG_FOLDER_VIEW,Flags) > 0;
end;

//****************************************************
function TNotesView.GetIsShared: Boolean;
begin
 Result := Pos(DESIGN_FLAG_PRIVATE_IN_DB,Flags) = 0;
end;

//****************************************************
class function TNotesView.GetViewID(notesDatabase: TNotesDatabase;
                  aName: string): NOTEID;
var
 ansiName2: AnsiString;
 err: DWORDapi;
 n: integer;
begin
 ansiName2 := Native2Lmbcs(AnsiString(aName));
 err := NIFFindDesignNote(notesDatabase.Handle, PAnsiChar(ansiName2), NOTE_CLASS_VIEW, @Result);
 if err <> 0 then
   err := NIFFindPrivateDesignNote(notesDatabase.Handle, PAnsiChar(ansiName2),
                               NOTE_CLASS_VIEW, @Result);
 if err <> 0 then begin
   // Do the same for view alias (if present)
   n := LastDelimiter('|',aName);
   if n > 0 then begin
     ansiName2 := AnsiString(copy(aName,n+1,length(aName)-n));
     ansiName2 := Native2Lmbcs(ansiName2);
     err := NIFFindDesignNote(notesDatabase.Handle, PAnsiChar(ansiName2),
                               NOTE_CLASS_VIEW, @Result);
     if err <> 0 then
       err := NIFFindPrivateDesignNote(notesDatabase.Handle, PAnsiChar(ansiName2),
                               NOTE_CLASS_VIEW, @Result);
   end;
 end;
 CheckError(err);
end;

//****************************************************
function TNotesView.GetOriginatorID: OID;
var
  td: TIMEDATE;
  cls: word;
begin
  CheckError(NSFDbGetNoteInfo(Database.Handle, ID, @Result, @td, @cls));
end;

//****************************************************
function TNotesView.GetUniversalID: UNID;
var
  id: OID;
begin
  id := GetOriginatorID;
  Result.aFile := id.FileNum;
  Result.Note := id.Note;
end;

//****************************************************
procedure TNotesView.Load(aKey: string; nMaxDocs: integer);
begin
 LoadByID(GetViewID(Database,Name), aKey, nMaxDocs);
end;

//****************************************************
procedure TNotesView.LoadByID(anID: NOTEID; aKey: string; nMaxDocs: integer);
var
 hColl:         HCOLLECTION;
 CollPosition:  COLLECTIONPOSITION;
 hBuffer:       LHandle;
 NotesFound:    DWORDapi;
 SignalFlags:   word;
 SkipNavigator: word;
 ReturnCount:   longint;
 Error:         word;
begin
 // Get flags
 FNoteID := anID;
 FKey    := aKey;
 FMaxDocs := nMaxDocs;
 Flags := LoadFlags (Database, ID);
 if (not IsFolder) and (self.className = TNotesFolder.className) then
   raise ELnInvalidFolder.Create ('Attempting to open view ' + Name + ' as a folder');

 // Load docs
 hColl := 0;
 try
   CheckError(NIFOpenCollection(
             Database.Handle,         { database handle }
             Database.Handle,         { database handle }
             ID,                      { view note ID}
             0,
             0,
             hColl,
             nil,
             nil,
             nil,
             nil));
   FillChar(CollPosition,sizeOf(CollPosition),0);

   // Added by piton
   if aKey = '' then begin
     SkipNavigator := NAVIGATE_NEXT;
     if nMaxDocs > 0 then
       ReturnCount := nMaxDocs
     else
       ReturnCount := $FFFF;
    end
    else begin
      error := NIFFindByName(hColl, PAnsiChar(AnsiString(FKey)),
                       FIND_CASE_INSENSITIVE, @CollPosition, @ReturnCount );
      if (error and $3fff) = ERR_NOT_FOUND then
        ReturnCount := 0
      else
        CheckError(error);
      SkipNavigator := NAVIGATE_CURRENT;
      if nMaxDocs > 0 then
        ReturnCount := nMaxDocs;
    end;

   repeat
     hBuffer := 0;
     try
       CheckError(NIFReadEntries(
          hColl,            { handle to this collection }
          @CollPosition,    { where to start in collection }
          SkipNavigator,    { order to skip entries }
          1,                { number to skip }
          NAVIGATE_NEXT,    { order to use after skipping }
          ReturnCount,      { max return number }
          READ_MASK_NOTEID or READ_MASK_SUMMARY,      { info we want }
          @hBuffer,         { handle to info (return) }
          nil,              { length of buffer (return) }
          nil,              { entries skipped (return) }
          NotesFound,       { number of notes (return) }
          SignalFlags));    { share warning (return) }
       if (hBuffer<>0) then
         GetCollectionByHandle(hBuffer, NotesFound, True);
     finally
       if hBuffer <> 0 then
         OSMemFree(hBuffer);
     end;
   until (SignalFlags and SIGNAL_MORE_TO_DO) = 0;
 finally
   if (hColl <> 0) then
     CheckError(NIFCloseCollection (hColl));
 end;
end;

//******************************************************************************

class function TNotesView.LoadFlags(Database: TNotesDatabase; anID: NOTEID): string;
var
 hNote: LHandle;
 ansiRet: AnsiString;
begin
 CheckError( NsfNoteOpen(Database.Handle, anID, OPEN_NOUPDATE, @hNote) );
 try
   ansiRet := '';
   setLength(ansiRet,31);
   NsfItemGetText(hNote, DESIGN_FLAGS, PAnsiChar(ansiRet), 30);
   ansiRet := System.AnsiStrings.StrPas(PAnsiChar(ansiRet));
 finally
   NsfNoteClose(hNote);
 end;
 Result := String(ansiRet);
end;

//******************************************************************************

class function TNotesView.OpenView(notesDatabase: TNotesDatabase;
                   aName: string): TNotesView;
begin
 Result := OpenViewExt(notesDatabase, aName, '', 0);
end;

class function TNotesView.OpenViewExt(notesDatabase: TNotesDatabase; aName,
                   aKey: string;
                   nMaxDocs: integer): TNotesView;
var
 Flags: string;
 ID: NOTEID;
begin
 Id := GetViewID(notesDatabase, aName);
 Flags := LoadFlags(notesDatabase, ID);
 if Pos(DESIGN_FLAG_FOLDER_VIEW,Flags) > 0 then
   Result := TNotesFolder.CreateEmpty(notesDatabase,aName)
 else
   Result := TNotesView.CreateEmpty(notesDatabase,aName);
 Result.LoadByID(ID, aKey, nMaxDocs);
end;

//******************************************************************************

procedure TNotesView.Update;
var
 hColl: HCOLLECTION;
begin
 CheckError( NIFOpenCollection(Database.Handle,Database.Handle, ID, 0, 0, hColl,
                           nil, nil, nil, nil));
 try
   CheckError( NIFUpdateCollection(hColl) );
   Clear;
   LoadByID(ID, Key, MaxDocs);
 finally
   NIFCloseCollection(hColl);
 end;
end;

//ClassMarker_Method(TNotesView)



//******************************************************************************
// TNotesFolder
//******************************************************************************


constructor TNotesFolder.CreateNew(notesDatabase: TNotesDatabase;
                                   aName:   string;
                                   fShared: boolean;
                                   FormatFolder: TNotesFolder);
var
 hFormatNote, hFormatDB: LHandle;
 dwDesign: DESIGN_TYPE;
 ansiName2: AnsiString;
begin
 CreateEmpty(notesDatabase, aName);
 Flags := DESIGN_FLAG_FOLDER_VIEW;
 if not fShared then
   Flags := Flags + DESIGN_FLAG_PRIVATE_IN_DB;
 ansiName2 := Native2Lmbcs(AnsiString(aName));
 if FormatFolder = nil then begin
   hFormatNote := 0;
   hFormatDB := 0;
 end
 else begin
   hFormatNote := FormatFolder.ID;
   hFormatDB   := FormatFolder.Database.Handle;
 end;

 if fShared then
   dwDesign := DESIGN_TYPE_SHARED
 else
   dwDesign := DESIGN_TYPE_PRIVATE_DATABASE;
 CheckError( FolderCreate(Database.Handle,0,hFormatNote,hFormatDB,PAnsiChar(ansiName2),
                          length(ansiName2),dwDesign,0,FNoteID) );
 Database.UpdateViews;
end;

//******************************************************************************

procedure TNotesFolder.AddDocument(Doc: TNotesDocument);
var
 Lst: TList;
begin
 Lst := TList.create;
 try
   Lst.add(Doc);
   AddDocuments(Lst);
 finally
   Lst.free;
 end;
end;

//******************************************************************************

procedure TNotesFolder.AddDocuments(DocList: TList);
var
 hTable: LHandle;
 i: integer;
begin
 CheckError( IDCreateTable(sizeOf(NOTEID),@hTable) );
 try
   for i := 0 to DocList.count-1 do
     CheckError( IDInsert(hTable,TNotesDocument(DocList[i]).DocID, nil) );
   CheckError( FolderDocAdd(Database.Handle,0,ID,hTable,0) );
 finally
   IDDestroyTable(hTable);
 end;
end;

//******************************************************************************

procedure TNotesFolder.RemoveDocuments;
begin
 CheckError( FolderDocRemoveAll(Database.Handle, 0, ID, 0) );
end;

//******************************************************************************

function TNotesFolder.Copy(NewName: string): TNotesFolder;
var
 nid: NOTEID;
 ansiName2: AnsiString;
begin
 ansiName2 := Native2Lmbcs(AnsiString(NewName));
 CheckError(FolderCopy(Database.Handle,0,ID,PAnsiChar(ansiName2),length(ansiName2),0,nid));
 Database.UpdateViews;
 Result := TNotesFolder.Create(Database,NewName);
end;

//******************************************************************************

procedure TNotesFolder.Delete;
begin
 CheckError(FolderDelete(Database.Handle,0,ID,0));
 Database.UpdateViews;
end;

//******************************************************************************

procedure TNotesFolder.DeleteDocument(Doc: TNotesDocument);
var
 Lst: TList;
begin
 Lst := TList.create;
 try
   Lst.add (Doc);
   DeleteDocuments(Lst);
 finally
   Lst.free;
 end;
end;

//******************************************************************************

procedure TNotesFolder.DeleteDocuments(DocList: TList);
var
 hTable: LHandle;
 i: integer;
 l_oNotesDocument: TNotesDocument;
 l_iID: NOTEID;
begin
 CheckError( IDCreateTable(sizeOf(NOTEID),@hTable) );
 try
   for i := 0 to DocList.count-1 do begin
     l_oNotesDocument := DocList[i];
     CheckError( IDInsert(hTable,l_oNotesDocument.DocID, nil) );
   end;
   l_iID := ID;
   CheckError( FolderDocRemove(Database.Handle,0,ID,hTable,0) );
 finally
   IDDestroyTable(hTable);
 end;
end;

//******************************************************************************

procedure TNotesFolder.Move(ParentFolder: TNotesFolder);
begin
 CheckError(FolderMove(Database.Handle,0,ID,0,ParentFolder.ID,0));
 Database.UpdateViews;
end;

//******************************************************************************

procedure TNotesFolder.SetName(Value: String);
var
 ansiName2: AnsiString;
begin
 if (Value <> FName) then begin
   ansiName2 := Native2Lmbcs(AnsiSTring(Value));
   CheckError(FolderRename(Database.Handle,0,ID,PAnsiChar(ansiName2),length(ansiName2),0));
   FName := Value;
   Database.UpdateViews;
 end;
end;
//ClassMarker_Method(TNotesFolder)


//******************************************************************************
// TAclRoles
//******************************************************************************

type
  TAclRoles = class(TStringList)
  private
    ACL: TNotesACL;
    InInit: boolean;
  protected
    procedure Put(Index: Integer; const S: string); override;
    procedure Init;
  public
    function  Add(const S: string): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
  end;

//******************************************************************************

procedure TAclRoles.Init;
var
 i: word;
 ansiBuf: AnsiString;
 Str: string;
 err: STATUS;
begin
 InInit := True;
 try
   Clear;
   Duplicates := dupIgnore;
   ansiBuf := '';
   SetLength(ansiBuf, ACL_PRIVSTRINGMAX+1);
   for i := 0 to ACL_PRIVCOUNT-1 do begin
     err := ACLGetPrivName(ACL.Handle,i,PAnsiChar(ansiBuf));
     if err <> 1060 then
       CheckError(err)
     else
       break;
     Str := trim(String(System.AnsiStrings.StrPas(PAnsiChar(ansiBuf))));
     if str <> '' then
       Add(Str);
   end;
 finally
   InInit := False;
 end;
end;

//******************************************************************************

procedure TAclRoles.Put(Index: Integer; const S: string);
begin
 if not InInit then
   CheckError( ACLSetPrivName(ACL.Handle, Index, PAnsiChar(AnsiString(S))) );
 inherited Put(Index,S);
end;

//******************************************************************************

function TAclRoles.Add(const S: string): Integer;
begin
 if not InInit then
   CheckError(ACLSetPrivName(ACL.Handle, Count+1, PAnsiChar(AnsiString(S))));
 Result := inherited Add(S);
end;

//******************************************************************************

procedure TAclRoles.Clear;
begin
 if InInit then
   inherited Clear
 else
   raise ELotusNotes.createErr(-1, 'Action is not supported');
end;

//******************************************************************************

procedure TAclRoles.Delete(Index: Integer);
begin
 raise ELotusNotes.createErr(-1, 'Action is not supported');
 //inherited Delete(Index);
end;

//******************************************************************************

procedure TAclRoles.Insert(Index: Integer; const S: string);
begin
 if not InInit then
   Add(S)
 else
   inherited Insert(Index,S);
end;

//******************************************************************************
// ACL maps
//******************************************************************************

const
  AccLevelMap: array [TNotesAclAccessLevel] of word = (
    ACL_LEVEL_NOACCESS,
    ACL_LEVEL_DEPOSITOR,
    ACL_LEVEL_READER,
    ACL_LEVEL_AUTHOR,
    ACL_LEVEL_EDITOR,
    ACL_LEVEL_DESIGNER,
    ACL_LEVEL_MANAGER
  );
  AclFlagMap: array [TNotesAclFlag] of word = (
    ACL_FLAG_AUTHOR_NOCREATE,
    ACL_FLAG_SERVER,
    ACL_FLAG_NODELETE,
    ACL_FLAG_CREATE_PRAGENT,
    ACL_FLAG_CREATE_PRFOLDER,
    ACL_FLAG_PERSON,
    ACL_FLAG_GROUP,
    ACL_FLAG_CREATE_FOLDER,
    ACL_FLAG_CREATE_LOTUSSCRIPT,
    ACL_FLAG_PUBLICREADER,
    ACL_FLAG_PUBLICWRITER,
    ACL_FLAG_ADMIN_READERAUTHOR,
    ACL_FLAG_ADMIN_SERVER
  );


//******************************************************************************
// TNotesACL
//******************************************************************************

constructor TNotesACL.Create(aDatabase: TNotesDatabase);
begin
 inherited Create;
 FDatabase := aDatabase;
 FRoles := TAclRoles.create;
 TAclRoles(FRoles).ACL := self;
 FEntries := TList.create;
 CheckError( NSFDBReadACL(Database.Handle,@FHandle) );
 TAclRoles(FRoles).Init;
 ReadEntries;
end;

//****************************************************
procedure TNotesACL.DeleteACLEntry(aName: String);
var
 e: TNotesAclEntry;
begin
 e := Entry[aName];
 if e = nil then
   exit;
 CheckError( AclDeleteEntry(Handle, PAnsiChar(AnsiString(aName))) );
 FEntries.Remove(e);
 e.free;
end;

//******************************************************************************

destructor TNotesACL.Destroy;
var
 i: integer;
begin
 for i := 0 to FEntries.count-1 do
   TObject(FEntries[i]).free;
 FEntries.free;
 FRoles.free;
 if FHandle <> 0 then
   OsMemFree(FHandle);
 inherited Destroy;
end;

//******************************************************************************

function TNotesACL.GetEntriesCount: Integer;
begin
 Result := FEntries.count;
end;

//******************************************************************************

function TNotesACL.GetEntry(aName: String): TNotesACLEntry;
var
 i: integer;
begin
 for i := 0 to EntriesCount-1 do begin
   Result := EntryByIndex[i];
   if compareText(Result.Name, aName) = 0 then
     exit;
 end;
 Result := nil;
end;

//******************************************************************************

function TNotesACL.GetEntryByIndex(Index: integer): TNotesACLEntry;
begin
 Result := TNotesAclEntry(FEntries[Index]);
end;

//******************************************************************************

function TNotesACL.GetUniformAccess: Boolean;
var
 f: DWORDapi;
begin
 CheckError( AclGetFlags(Handle, f) );
 Result := (ACL_UNIFORM_ACCESS and f) <> 0;
end;

//******************************************************************************

procedure TNotesACL.SetUniformAccess(Value: boolean);
var
 f: DWORDapi;
begin
 if Value then
   f := ACL_UNIFORM_ACCESS
 else
   f := 0;
 CheckError( AclSetFlags(Handle, f) );
end;

//******************************************************************************

function TNotesACL.CreateACLEntry(aName: string): TNotesACLEntry;
begin
 if Entry[aName] <> nil then
   raise ELotusNotes.CreateErr(-1, 'Entry with name ' + aName + ' already exists');
 Result := TNotesAclEntry.CreateNew(self, aName, aclNoAccess, '');
 FEntries.Add(Result);
end;

//******************************************************************************

procedure EntriesProc (Param: pointer; Name: PAnsiChar; AccLevel: word;
                       Privileges: PACL_PRIVILEGES; AccFlags: WORD); stdcall; far;
var
 e: TNotesACLEntry;
 AccessLevel: TNotesAclAccessLevel;
 AccessFlags: TNotesAclFlags;
 Flag: TNotesAclFlag;
begin
 for AccessLevel := System.Low(TNotesAclAccessLevel) to System.High(TNotesAclAccessLevel) do
   if AccLevelMap[AccessLevel] = AccLevel then break;
 AccessFlags := [];
 for Flag := System.Low(TNotesAclFlag) to System.High(TNotesAclFlag) do
    if (AccFlags and AclFlagMap[Flag]) <> 0 then include(AccessFlags, Flag);
 e := TNotesACLEntry.create(TNotesAcl(Param), String(System.AnsiStrings.StrPas(Name)),
               AccessLevel, Privileges^, AccessFlags);
 TNotesAcl(Param).FEntries.Add(e);
end;

//******************************************************************************

procedure TNotesACL.ReadEntries;
begin
  CheckError( AclEnumEntries(Handle,EntriesProc,self) );
end;

//******************************************************************************

procedure TNotesACL.Save;
var
 i: integer;
begin
 for i := 0 to EntriesCount-1 do
   EntryByIndex[i].Update;
 CheckError( NsfDBStoreAcl(Database.Handle, FHandle, 0, 0) );
end;
//ClassMarker_Method(TNotesACL)



//******************************************************************************
// TNotesACLEntry
//******************************************************************************

procedure TNotesACLEntry.AddRole(aRole: string);
var
 s: string;
begin
 s := Roles;
 if Pos(aRole+#13#10, s) = 0 then
   Roles := s + aRole + #13#10;
end;

//******************************************************************************

constructor TNotesACLEntry.Create(anACL: TNotesACL; aName: string; AccLevel: TNotesAclAccessLevel;
      const AclPrivs: ACL_PRIVILEGES; AccFlags: TNotesAclFlags);
begin
 inherited Create;
 FUpdateFlags := 0;
 FAcl := anAcl;
 FName := aName;
 FAccessLevel := AccLevel;
 FPrivileges := AclPrivs;
 FFlags := AccFlags;
end;

//******************************************************************************

constructor TNotesACLEntry.CreateNew(anACL: TNotesACL;
                             aName: string;
                             AccLevel: TNotesAclAccessLevel;
                             aRoles: string);
begin
 inherited Create;
 FNew := True;
 FUpdateFlags := ACL_UPDATE_NAME or ACL_UPDATE_LEVEL or ACL_UPDATE_PRIVILEGES or ACL_UPDATE_FLAGS;
 FAcl := anAcl;
 Name := aName;
 AccessLevel := AccLevel;
 Roles := aRoles;
end;

//******************************************************************************

procedure TNotesACLEntry.SetAccessLevel(Value: TNotesAclAccessLevel);
begin
 if (FAccessLevel <> Value) then begin
   FAccessLevel := Value;
   FUpdateFlags := FUpdateFlags or ACL_UPDATE_LEVEL;
 end;
end;

//******************************************************************************

procedure TNotesACLEntry.SetFlags(Value: TNotesAclFlags);
begin
 if (FFlags <> Value) then begin
   FFlags := Value;
   FUpdateFlags := FUpdateFlags or ACL_UPDATE_FLAGS;
 end;
end;

//******************************************************************************
procedure TNotesACLEntry.SetName(Value: String);
begin
 if (FName <> Value) then begin
   if FOldName = '' then
     FOldName := FName;
   FName := Value;
   FUpdateFlags := FUpdateFlags or ACL_UPDATE_NAME;
 end;
end;

//******************************************************************************

procedure TNotesACLEntry.SetPrivileges(Value: ACL_PRIVILEGES);
begin
 FPrivileges := Value;
 FUpdateFlags := FUpdateFlags or ACL_UPDATE_PRIVILEGES;
end;

//******************************************************************************

procedure TNotesACLEntry.SetRoles(Value: string);
var
 rl: TStringList;
 i, n: integer;
begin
 rl := TStringList.create;
 try
   rl.Text := Value;
   FillChar (FPrivileges, sizeOf(FPrivileges), 0);
   for i := 0 to rl.count-1 do begin
     n := ACL.Roles.IndexOf(rl[i]);
     if n >= 0 then
       AclSetPriv (FPrivileges, n + ACL_BITPRIVCOUNT);
   end;
   FUpdateFlags := FUpdateFlags or ACL_UPDATE_PRIVILEGES;
 finally
   rl.free;
 end;
end;

//******************************************************************************

procedure TNotesACLEntry.DeleteRole(aRole: string);
var
 s: string;
 n: integer;
begin
 s := Roles;
 n := Pos (aRole+#13#10, s);
 if n > 0 then begin
   delete (s, n, length(aRole + #13#10));
   Roles := s;
 end;
end;

//******************************************************************************

destructor TNotesACLEntry.Destroy;
begin
  Update;
  inherited Destroy;
end;

//******************************************************************************

function TNotesACLEntry.GetRoles: String;
var
 i: integer;
begin
 Result := '';
 for i := 0 to ACL_PRIVCOUNT-1 do if ACLIsPrivSet(FPrivileges, i) then begin
   Result := Result + ACL.Roles[i-ACL_BITPRIVCOUNT] + #13#10;
 end;
end;

//******************************************************************************

procedure TNotesACLEntry.Update;
var
 flag: TNotesAclFlag;
 wFlags: word;
 pcName, pcOldName: PAnsiChar;
begin
 if (not FNew) and (FUpdateFlags = 0) then
   exit;
 if FOldName = '' then
   FOldName := FName;
 // By Joe Barbaretta
 // Default ACL entry is represented by NIL
 if (FName = '') or (compareText(FName,'-Default-') = 0) then
   pcName := nil
 else
   pcName := PAnsiChar(AnsiString(FName));
 if (FOldName = '') or (compareText(FOldName,'-Default-') = 0) then
   pcOldName := nil
 else
   pcOldName := PAnsiChar(AnsiString(FOldName));

 wFlags := 0;
 for flag := System.Low(TNotesAclFlag) to System.High(TNotesAclFlag) do
   if flag in Flags then wFlags := wFlags or AclFlagMap[flag];
 if FNew then
   CheckError( AclAddEntry(ACL.Handle, pcName, AccLevelMap[AccessLevel], FPrivileges, wFlags) )
 else
   CheckError( AclUpdateEntry(Acl.Handle,
                  pcOldName,
                  FUpdateFlags,
                  pcName,
                  AccLevelMap[AccessLevel],
                  FPrivileges,
                  wFlags));
 FUpdateFlags := 0;
 FNew := False;
 FOldName := '';
end;


{$IFNDEF NO_INIT_SECTION}
initialization
  CheckError (NotesInit);
{$IFNDEF NO_CTRL_BREAK}
  OldBreakProc := OsGetSignalHandler (OS_SIGNAL_CHECK_BREAK);
  OsSetSignalHandler (OS_SIGNAL_CHECK_BREAK, @NDBreakProc);
{$ENDIF}
finalization
{$IFNDEF NO_CTRL_BREAK}
  OsSetSignalHandler (OS_SIGNAL_CHECK_BREAK, @OldBreakProc);
{$ENDIF}
  NotesTerm;
{$ENDIF}
end.



